# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for tuned_gp_models."""

import functools

from absl import logging
import jax
from jax.config import config
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import gp_bandit_utils
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import tuned_gp_models
from vizier._src.jax.optimizers import optimizers

from absl.testing import absltest

tfb = tfp.bijectors


class VizierGpTest(absltest.TestCase):

  def _generate_xys(self):
    x_obs = np.array(
        [
            [
                0.2941264,
                0.29313548,
                0.68817519,
                0.37502566,
                0.48356813,
                0.34127283,
            ],
            [
                0.66218224,
                0.70770083,
                0.6901334,
                0.66787973,
                0.5400858,
                0.52721233,
            ],
            [
                0.88469647,
                0.50593371,
                0.83160862,
                0.58674892,
                0.42145673,
                0.31749428,
            ],
            [
                0.39976682,
                0.59517741,
                0.73295106,
                0.6084903,
                0.54891015,
                0.44338632,
            ],
            [
                0.8354305,
                0.87605574,
                0.47855956,
                0.48174861,
                0.37685449,
                0.38348768,
            ],
            [
                0.55608455,
                0.72781129,
                0.52432913,
                0.44291417,
                0.3816395,
                0.326599,
            ],
            [
                0.24689187,
                0.50979672,
                0.67604857,
                0.45172594,
                0.34994392,
                0.75239792,
            ],
            [
                0.71007257,
                0.60896354,
                0.29270877,
                0.74683367,
                0.50169051,
                0.74480515,
            ],
            [
                0.9193235,
                0.24393112,
                0.63868591,
                0.43271524,
                0.43339578,
                0.59413154,
            ],
            [
                0.51850627,
                0.62689204,
                0.76134879,
                0.65990021,
                0.82350868,
                0.7429215,
            ],
        ],
        dtype=np.float64,
    )
    y_obs = np.array(
        [
            0.55552674,
            -0.29054829,
            -0.04703586,
            0.0217839,
            0.15445438,
            0.46654119,
            0.12255823,
            -0.19540335,
            -0.11772564,
            -0.44447326,
        ],
        dtype=np.float64,
    )
    return x_obs, y_obs

  def test_masking_works(self):
    # Mask three dimensions and four observations.
    observation_is_missing = np.array(
        [False, False, True, False, True, False, True, True, False, False]
    )
    dimension_is_missing = np.array([False, True, True, False, True, False])
    x_obs, y_obs = self._generate_xys()

    # Change these to nans to ensure that they are ignored.
    modified_x_obs = np.where(dimension_is_missing, np.nan, x_obs)
    modified_y_obs = np.where(observation_is_missing, np.nan, y_obs)

    data = types.StochasticProcessModelData(
        features=x_obs,
        labels=y_obs,
        dimension_is_missing=dimension_is_missing,
        label_is_missing=observation_is_missing,
    )
    model1 = tuned_gp_models.VizierGaussianProcess.build_model(features=x_obs)
    loss_fn1 = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model1,
        data=data,
    )

    modified_data = types.StochasticProcessModelData(
        features=modified_x_obs,
        labels=modified_y_obs,
        dimension_is_missing=dimension_is_missing,
        label_is_missing=observation_is_missing,
    )
    model2 = tuned_gp_models.VizierGaussianProcess.build_model(
        features=modified_x_obs
    )
    loss_fn2 = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model2,
        data=modified_data,
    )

    # Check that the model loss and optimal parameters are independent of those
    # dimensions and observations.
    optimize = optimizers.JaxoptLbfgsB(random_restarts=1)
    optimal_params1, _ = optimize(
        lambda rng: model1.init(rng, x_obs)['params'],
        loss_fn1,
        jax.random.PRNGKey(2),
        constraints=sp.get_constraints(model1),
    )
    optimal_params2, _ = optimize(
        lambda rng: model2.init(rng, modified_x_obs)['params'],
        loss_fn2,
        jax.random.PRNGKey(2),
        constraints=sp.get_constraints(model2),
    )

    for key in optimal_params1:
      self.assertTrue(
          np.all(np.equal(optimal_params1[key], optimal_params2[key])),
          msg=f'{key} parameters were not equal.',
      )
    self.assertEqual(loss_fn1(optimal_params1)[0], loss_fn2(optimal_params2)[0])

  def test_good_log_likelihood(self):
    x_obs, y_obs = self._generate_xys()
    target_loss = -0.2
    data = types.StochasticProcessModelData(features=x_obs, labels=y_obs)
    model = tuned_gp_models.VizierGaussianProcess.build_model(features=x_obs)
    loss_fn = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model,
        data=data,
    )
    setup = lambda rng: model.init(rng, x_obs)['params']

    optimize = optimizers.JaxoptLbfgsB(random_restarts=50)
    constraints = sp.get_constraints(model)
    optimal_params, metrics = optimize(
        setup, loss_fn, jax.random.PRNGKey(2), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)
    logging.info('Loss: %s', loss_fn(optimal_params)[0])
    self.assertLess(np.min(metrics['loss']), target_loss)

  def test_good_log_likelihood_with_masks(self):
    x_obs, y_obs = self._generate_xys()
    # Pad x_s and y_s and generate masks.
    x_obs = np.pad(x_obs, ((0, 2), (0, 3)), constant_values=np.nan)
    y_obs = np.pad(y_obs, (0, 2), constant_values=-np.inf)

    observation_is_missing = np.array(
        [False] * (y_obs.shape[0] - 2) + [True] * 2
    )
    dimension_is_missing = np.array(
        [False] * (x_obs.shape[-1] - 3) + [True] * 3
    )

    data = types.StochasticProcessModelData(
        features=x_obs,
        labels=y_obs,
        dimension_is_missing=dimension_is_missing,
        label_is_missing=observation_is_missing,
    )
    target_loss = -0.2
    model = tuned_gp_models.VizierGaussianProcess.build_model(features=x_obs)
    loss_fn = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model,
        data=data,
    )
    setup = lambda rng: model.init(rng, x_obs)['params']

    optimize = optimizers.JaxoptLbfgsB(random_restarts=50)
    constraints = sp.get_constraints(model)
    optimal_params, metrics = optimize(
        setup, loss_fn, jax.random.PRNGKey(2), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)
    logging.info('Loss: %s', loss_fn(optimal_params)[0])
    self.assertLess(np.min(metrics['loss']), target_loss)

  def test_good_log_likelihood_with_masks_continuous_and_categorical(self):
    x_cont_obs, y_obs = self._generate_xys()
    x_cat_obs = np.int_(self._generate_xys()[0] * 10)
    x_obs = types.ContinuousAndCategoricalArray(
        continuous=x_cont_obs, categorical=x_cat_obs
    )
    target_loss = -0.2
    model = tuned_gp_models.VizierGaussianProcessWithCategorical.build_model(
        x_obs,
    )
    setup = lambda rng: model.init(rng, x_obs)['params']
    data = types.StochasticProcessModelData(features=x_obs, labels=y_obs)
    loss_fn = functools.partial(
        jax.jit(
            gp_bandit_utils.stochastic_process_model_loss_fn,
            static_argnames=('model', 'normalize'),
        ),
        model=model,
        data=data,
    )
    optimize = optimizers.JaxoptLbfgsB(random_restarts=50)
    constraints = sp.get_constraints(model)
    optimal_params, metrics = optimize(
        setup, loss_fn, jax.random.PRNGKey(2), constraints=constraints
    )
    logging.info('Optimal: %s', optimal_params)
    logging.info('Loss: %s', loss_fn(optimal_params)[0])
    self.assertLess(np.min(metrics['loss']), target_loss)


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Fast pareto frontier computation using Jax."""
import functools
from typing import Any

import jax
from jax import numpy as jnp
import numpy as np

from vizier._src.pyvizier.multimetric import pareto_optimal

# Type aliases for readability only.
# We don't over-specify for compatibility between jax and np.
Matrix = Any
Vector = Any
Scalar = Any


# y2 dominates y1 if y2 >= y1 coordinate wise. This is strict if we also have
# y2 > y1 in some coordinate.
def _jax_dominated(y1: Vector,
                   y2: Vector,
                   strict: bool = True) -> Scalar:  # ([a], [a]) -> []
  dominated_or_equal = jnp.all(y1 <= y2)
  if strict:
    return dominated_or_equal & jnp.any(y2 > y1)
  else:
    return dominated_or_equal


def _is_pareto_optimal_against(yy: Matrix, baseline: Matrix, *,
                               strict: bool) -> Vector:
  """Computes if nothing in `baseline` dominates `yy`.

  Args:
    yy: array of shape [N_y, M] where M is number of metrics.
    baseline: array of shape [N_b, M] where M is number of metrics.
    strict: If true, strict dominance is used.

  Returns:
    Boolean array of shape [N_y]
  """
  jax_dominated_mv = jax.vmap(
      functools.partial(_jax_dominated, strict=strict), (None, 0),
      0)  #  ([b,a], [a]) -> [b]
  jax_dominated_mm = jax.vmap(jax_dominated_mv, (0, None),
                              0)  #  ([b,a], [c,a]) -> [b,c]
  return jnp.logical_not(jnp.any(jax_dominated_mm(yy, baseline),
                                 axis=-1))  # N_y, N_b -> N_y


_jax_is_pareto_optimal_against = jax.jit(
    _is_pareto_optimal_against, static_argnames='strict')


def is_frontier(ys: Matrix,
                *,
                num_shards: int = 10,
                verbose: bool = False) -> jnp.ndarray:
  """Efficiently compute `_jax_is_pareto_optimal_against(ys, ys, strict=True)`.

  Divide `ys` into shards and gradually trim down the candidates.

  Args:
    ys: Array of shape [N_y, M] where M is number of metrics.
    num_shards: Each sharding results in filtering, i.e. indexing the array with
      boolean vector. This operation can be very expensive and dominate the cost
      of computation. Use a moderate number sublinear in N_y, e.g. log(N).
    verbose:

  Returns:
    Boolean Array of shape [N_y].
  """
  idx = np.linspace(0, ys.shape[0], num_shards).astype(np.int32)
  idx = list(reversed(idx))
  # Initialize candidates with all points
  frontier = np.ones(ys.shape[0], dtype=np.bool_)

  for begin, end in zip(idx[1:], idx[:-1]):
    candidates = ys[frontier]
    # Filter candidates by comparing against the slice.
    if verbose:
      print(f'Compare {len(candidates)} against {begin}:{end}.')
    tt = _jax_is_pareto_optimal_against(candidates, ys[begin:end], strict=True)
    frontier[frontier] = tt
  return frontier


class JaxParetoOptimalAlgorithm(pareto_optimal.BaseParetoOptimalAlgorithm):
  """Jax-based functions for calculating pareto frontiers."""

  def is_pareto_optimal(self, points: np.ndarray) -> np.ndarray:
    """Jax-enabled Pareto frontier algorithm. See base class."""
    return np.array(is_frontier(points), dtype=bool)

  def is_pareto_optimal_against(self, points: np.ndarray,
                                dominating_points: np.ndarray,
                                strict: bool) -> np.ndarray:
    """Jax-enabled optimality/domination algorithm. See base class."""
    return np.array(
        _jax_is_pareto_optimal_against(points, dominating_points, strict),
        dtype=bool)


def get_frontier(ys: Matrix,
                 *,
                 num_shards: int = 10,
                 verbose: bool = True) -> jnp.ndarray:
  """Efficiently compute `ys[_jax_is_pareto_optimal_against(ys, ys, strict=True)]` using iterative filtering.

  Divide `ys` into shards and gradually trim down the candidates.
  `get_frontier` doesn't call `is_frontier`, because `get_frontier` runs faster
  by not slicing the full `ys` every iteration.

  Args:
    ys: Array of shape [N_y, M] where M is number of metrics.
    num_shards: Each sharding results in filtering, i.e. indexing the array with
      boolean vector. This operation can be very expensive and dominate the cost
      of computation. Use a moderate number sublinear in N_y, e.g. log(N).
    verbose:

  Returns:
    Array of shape [N_y, M].
  """
  idx = np.linspace(0, ys.shape[0], num_shards).astype(np.int32)
  idx = list(reversed(idx))
  # Initialize candidates with all points
  candidates = jnp.asarray(ys)
  for begin, end in zip(idx[1:], idx[:-1]):
    # Filter candidates by comparing against the slice.
    if verbose:
      # Use print. This method won't run in production anyways.
      print(f'Compare {len(candidates)} against {begin}:{end}.')
    tt = _jax_is_pareto_optimal_against(candidates, ys[begin:end], strict=True)
    candidates = candidates[tt]
  return candidates


@jax.jit
def pareto_rank(ys: Matrix) -> jnp.ndarray:
  jax_dominated_mv = jax.vmap(
      functools.partial(_jax_dominated, strict=True), (None, 0),
      0)  #  ([b,a], [a]) -> [b]
  jax_dominated_mm = jax.vmap(jax_dominated_mv, (0, None),
                              0)  #  ([b,a], [c,a]) -> [b,c]
  domination_matrix = jax_dominated_mm(ys, ys)
  return jnp.sum(domination_matrix, axis=1)


def _cum_hypervolume_origin(points: Matrix, vector: Vector) -> Vector:
  """Returns a randomized approximation of the cumulative dominated hypervolume.

  See Section 3, Lemma 5 of https://arxiv.org/pdf/2006.04655.pdf for a fuller
  explanation of the technique. This assumes the reference point is the
  origin.

  NOTE: This returns an unnormalized hypervolume.

  Args:
    points: Any set of points with shape (num_points, dimension).
    vector: A vector of length dimension.

  Returns:
    Approximated cumulative dominated hypervolume of points[:i]. Length is
    num_points.
  """
  ratios = points / vector
  coordinate_min_ratio = jnp.min(ratios, axis=1)
  return jax.lax.cummax(coordinate_min_ratio, axis=0)**len(vector)


_cum_hypervolume_mm = jax.vmap(_cum_hypervolume_origin, (None, 0),
                               0)  #  ([b,a], [c,a]) -> [b,c]


@jax.jit
def jax_cum_hypervolume_origin(points, vectors):
  return jnp.mean(_cum_hypervolume_mm(points, vectors), axis=0)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gp_bandit."""

import numpy as np
from vizier.pyvizier.converters import padding

from absl.testing import absltest
from absl.testing import parameterized


class PaddingTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
          num_trials=17,
          num_features=5,
          expected_num_trials=17,
          expected_num_features=5,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
          num_trials=23,
          num_features=8,
          expected_num_trials=23,
          expected_num_features=8,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          num_trials=17,
          num_features=5,
          expected_num_trials=20,
          expected_num_features=10,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          num_trials=23,
          num_features=2,
          expected_num_trials=32,
          expected_num_features=10,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          num_trials=7,
          num_features=8,
          expected_num_trials=10,
          expected_num_features=10,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          num_trials=7,
          num_features=20,
          expected_num_trials=10,
          expected_num_features=20,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          num_trials=123,
          num_features=22,
          expected_num_trials=130,
          expected_num_features=32,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          num_trials=17,
          num_features=5,
          expected_num_trials=32,
          expected_num_features=8,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          num_trials=23,
          num_features=2,
          expected_num_trials=32,
          expected_num_features=2,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          num_trials=7,
          num_features=8,
          expected_num_trials=8,
          expected_num_features=10,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          num_trials=7,
          num_features=17,
          expected_num_trials=8,
          expected_num_features=32,
      ),
      dict(
          schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          num_trials=123,
          num_features=22,
          expected_num_trials=128,
          expected_num_features=32,
      ),
  )
  def test_padding(
      self,
      schedule,
      num_trials,
      num_features,
      expected_num_trials,
      expected_num_features,
  ):
    features = np.random.randn(num_trials, num_features)
    labels = np.random.randn(num_trials)[..., np.newaxis]

    padded_features, padded_labels = padding.pad_features_and_labels(
        features, labels, schedule
    )

    self.assertLen(padded_features.is_missing, 2)
    self.assertLen(padded_labels.is_missing, 1)

    self.assertTrue(
        np.all(
            np.isclose(
                padded_features.is_missing[0], padded_labels.is_missing[0]
            )
        )
    )

    label_is_missing = padded_features.is_missing[0]
    feature_is_missing = padded_features.is_missing[1]
    padded_features = padded_features.padded_array
    padded_labels = padded_labels.padded_array

    self.assertEqual(
        padded_features.shape, (expected_num_trials, expected_num_features)
    )
    self.assertEqual(feature_is_missing.shape, (expected_num_features,))
    self.assertEqual(padded_labels.shape, (expected_num_trials, 1))
    self.assertEqual(label_is_missing.shape, (expected_num_trials,))

    self.assertTrue(
        np.all(
            np.isclose(features, padded_features[:num_trials, :num_features])
        )
    )
    self.assertTrue(np.all(np.isclose(labels, padded_labels[:num_trials])))
    self.assertTrue(np.all(~label_is_missing[:num_trials]))
    self.assertTrue(np.all(label_is_missing[num_trials:]))

    self.assertTrue(np.all(~feature_is_missing[:num_features]))
    self.assertTrue(np.all(feature_is_missing[num_features:]))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for profiler."""
import jax
import jax.numpy as jnp

from vizier.utils import profiler
from absl.testing import absltest


class PerformanceUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    profiler.Storage().reset()

  def test_no_args(self):
    class TestClass:

      @profiler.record_runtime(also_log=True)
      def add_one(self, x: int) -> int:
        return x + 1

    t = TestClass()
    t.add_one(1)
    self.assertIn('add_one', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('add_one')
    self.assertLen(runtimes, 1)
    self.assertGreaterEqual(runtimes[0].total_seconds(), 0)

  def test_with_name(self):
    class TestClass3:

      @profiler.record_runtime(name='add')
      def add_one(self, x: int) -> int:
        return x + 1

    t = TestClass3()
    t.add_one(1)
    self.assertNotIn('add_one', profiler.Storage().runtimes())
    self.assertIn('add', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('add')
    self.assertLen(runtimes, 1)
    self.assertGreaterEqual(runtimes[0].total_seconds(), 0)

  def test_with_prefix(self):
    class TestClass4:

      @profiler.record_runtime(name_prefix='foo')
      def add_one(self, x: int) -> int:
        return x + 1

    t = TestClass4()
    t.add_one(1)
    self.assertIn('foo.add_one', profiler.Storage().runtimes())
    self.assertNotIn('add_one', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('foo.add_one')
    self.assertLen(runtimes, 1)
    self.assertGreaterEqual(runtimes[0].total_seconds(), 0)

  def test_with_prefix_and_name(self):
    class TestClass5:

      @profiler.record_runtime(name_prefix='foo', name='add')
      def add_one(self, x: int) -> int:
        return x + 1

    t = TestClass5()
    t.add_one(1)
    self.assertIn('foo.add', profiler.Storage().runtimes())
    self.assertNotIn('add_one', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('foo.add')
    self.assertLen(runtimes, 1)
    self.assertGreaterEqual(runtimes[0].total_seconds(), 0)

  def test_with_jax(self):
    @profiler.record_runtime(also_log=True)
    @jax.jit
    def selu(x, alpha=1.67, lambda_=1.05):
      return lambda_ * jnp.where(x > 0, x, alpha * jnp.exp(x) - alpha)

    x = jnp.arange(1000000)
    # The first call should jit compile, so should be slow.
    selu(x)
    self.assertIn('selu', profiler.Storage().runtimes())
    microseconds = profiler.Storage().runtimes().get('selu')[0].microseconds
    self.assertGreaterEqual(microseconds, 1000)
    # The second call should use the jitted function.
    selu(x)
    self.assertLen(profiler.Storage().runtimes().get('selu'), 2)
    jt_microseconds = profiler.Storage().runtimes().get('selu')[1].microseconds
    # Should be at least 7 times faster (rough estimate so it's not flaky).
    self.assertLess(jt_microseconds, microseconds / 7)

  def test_with_jax_reverse_decorator_order(self):
    # If jax.jit decorator is first, only the first runtime is recorded.
    @jax.jit
    @profiler.record_runtime(also_log=True)
    def selu(x, alpha=1.67, lambda_=1.05):
      return lambda_ * jnp.where(x > 0, x, alpha * jnp.exp(x) - alpha)

    x = jnp.arange(1000000)
    # The first call should jit compile, so should be slow.
    selu(x)
    self.assertIn('selu', profiler.Storage().runtimes())
    microseconds = profiler.Storage().runtimes().get('selu')[0].microseconds
    self.assertGreaterEqual(microseconds, 1000)
    # The second call should use the jitted function.
    selu(x)
    self.assertLen(profiler.Storage().runtimes().get('selu'), 1)
    jt_microseconds = profiler.Storage().runtimes().get('selu')[0].microseconds
    # Should be at least 10 times faster.
    self.assertEqual(jt_microseconds, microseconds)

  def test_context_manager(self):
    x = 0
    with profiler.record_runtime_context(name='add'):
      x += 1
    self.assertIn('add', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('add')
    self.assertLen(runtimes, 1)
    self.assertGreaterEqual(runtimes[0].total_seconds(), 0)

    with profiler.record_runtime_context(name='add'):
      x += 1
    runtimes = profiler.Storage().runtimes().get('add')
    self.assertLen(runtimes, 2)
    self.assertGreaterEqual(runtimes[1].total_seconds(), 0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for pythia."""

import functools
import pyglove as pg
from vizier._src.algorithms.testing import test_runners
from vizier._src.pyglove import converters
from vizier._src.pyglove import pythia as pg_pythia
from absl.testing import absltest


class DummyAlgorithm(pg.DNAGenerator):
  """Dummy algorithm for testing."""

  def __init__(self):
    super().__init__()
    self._random = pg.geno.Random(seed=1)

  def setup(self, dna_spec: pg.DNASpec):
    super().setup(dna_spec)
    self._random.setup(dna_spec)

  def _propose(self) -> pg.DNA:
    return self._random.propose()


class PythiaTest(absltest.TestCase):

  def test_simple_search_space(self):
    # Get a DNA spec.
    rewards = []

    def foo():
      r = pg.oneof([1, 2]) + pg.floatv(3., 4.)
      branch = pg.oneof(['negate', 'no-op'])
      if branch == 'negate':
        r = -r
      rewards.append(r)
      return r

    search_space = pg.hyper.DynamicEvaluationContext()
    with search_space.collect():
      foo()

    # Create a vizier converter.
    converter = converters.VizierConverter.from_dna_spec(
        search_space.dna_spec, ('',)
    )

    # Create pyglove algorithm.
    algorithm = DummyAlgorithm()
    algorithm.setup(search_space.dna_spec)
    policy_factory = functools.partial(
        pg_pythia.TunerPolicy,
        converter=converter,
        algorithm=algorithm,
    )
    # Test the policy.
    tester = test_runners.RandomMetricsRunner(
        converter.problem,
        batch_size=5,
        validate_parameters=True,
    )
    tester.run_policy(policy_factory)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Acquisition functions and builders implementations."""

from typing import Mapping, Optional, Protocol, Sequence

from flax import struct
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import types
from vizier.pyvizier import converters


tfd = tfp.distributions
tfp_bo = tfp.experimental.bayesopt
tfpke = tfp.experimental.psd_kernels


class AcquisitionFunction(Protocol):

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
  ) -> jax.Array:
    pass


# Vizier library acquisition functions use `flax.struct`, instead of `attrs` and
# a hash function, so that acquisition functions can be passed as args to JIT-ed
# functions without triggering retracing when attribute values change.
@struct.dataclass
class UCB(AcquisitionFunction):
  """UCB AcquisitionFunction."""

  coefficient: float = 1.8

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features, labels, seed
    return dist.mean() + self.coefficient * dist.stddev()


@struct.dataclass
class HyperVolumeScalarization(AcquisitionFunction):
  """HyperVolume Scalarization acquisition function."""

  coefficient: float = 1.0

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features, labels, seed
    # Uses scalarizations in https://arxiv.org/abs/2006.04655 for
    # non-convex biobjective optimization of mean vs stddev.
    return jnp.minimum(dist.mean(), self.coefficient * dist.stddev())


@struct.dataclass
class EI(AcquisitionFunction):

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features, seed
    return tfp_bo.acquisition.GaussianProcessExpectedImprovement(dist, labels)()


@struct.dataclass
class PI(AcquisitionFunction):

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features, seed
    return tfp_bo.acquisition.GaussianProcessProbabilityOfImprovement(
        dist, labels
    )()


@struct.dataclass
class QEI(AcquisitionFunction):
  """Sampling-based batch expected improvement."""

  num_samples: int = 100

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features
    if seed is None:
      raise ValueError('QEI requires a value for `seed`.')
    return tfp_bo.acquisition.ParallelExpectedImprovement(
        dist, labels, seed=seed, num_samples=self.num_samples
    )()


@struct.dataclass
class QUCB(AcquisitionFunction):
  """Sampling-based batch upper confidence bound.

  Attributes:
    coefficient: UCB coefficient. For a Gaussian distribution, note that
      `UCB(coefficient=c)` is equivalent to `QUCB(coefficient=c * sqrt(pi / 2))`
      if QUCB batch size is 1. See the TensorFlow Probability docs for more
      details:
      https://www.tensorflow.org/probability/api_docs/python/tfp/experimental/bayesopt/acquisition/ParallelUpperConfidenceBound
    num_samples: Number of distribution samples used to compute qUCB.
    seed: Random seed for sampling.
  """

  coefficient: float = 1.8
  num_samples: int = 100

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    del features
    if seed is None:
      raise ValueError('QEI requires a value for `seed`.')
    return tfp_bo.acquisition.ParallelUpperConfidenceBound(
        dist,
        labels,
        seed=seed,
        exploration=self.coefficient,
        num_samples=self.num_samples,
    )()


@struct.dataclass
class MultiAcquisitionFunction(AcquisitionFunction):
  """Wrapper that calls multiple acquisition functions."""

  acquisition_fns: Mapping[str, AcquisitionFunction]

  def __call__(
      self,
      dist: tfd.Distribution,
      features: Optional[types.Features] = None,
      labels: Optional[types.Array] = None,
      seed: Optional[jax.random.KeyArray] = None,
  ) -> jax.Array:
    acquisitions = []
    for acquisition_fn in self.acquisition_fns.values():
      acquisitions.append(acquisition_fn(dist, features, labels))
    # TODO: Change the return type to a dict with the same
    # structure as `acquisition_fns` to clarify the meaning of the return
    # values.
    return jnp.stack(acquisitions, axis=0)


# TODO: Support discretes and categoricals.
# TODO: Support custom distances.
@struct.dataclass
class TrustRegion:
  """L-inf norm based TrustRegion.

  Limits the suggestion within the union of small L-inf norm balls around each
  of the trusted points, which are in most cases observed points. The radius
  of the L-inf norm ball grows in the number of observed points.

  Assumes that all points are in the unit hypercube.

  The trust region can be used e.g. during acquisition optimization:
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    features, labels = converter.to_xy(trials)
    tr = TrustRegion(features, converter.output_specs)
    # xs is a point in the search space.
    distance = tr.min_linf_distance(xs)
    if distance <= tr.trust_radius:
      print('xs in trust region')
  """

  max_distances: types.Array
  data: types.StochasticProcessModelData
  trust_radius: types.Array

  @classmethod
  def build(
      cls,
      specs: Optional[Sequence[converters.NumpyArraySpec]],
      data: types.StochasticProcessModelData,
  ) -> 'TrustRegion':
    """Init.

    Args:
      specs: List of output specs of the `TrialToArrayConverter`.
      data: Structure of observed features and labels, with optional boolean
        masks.

    Returns:
      trust_region
    """

    # TODO: Make hyperparameters configurable.
    min_radius = 0.2  # Hyperparameter
    dimension_factor = 5.0  # Hyperparameter

    if isinstance(data.features, types.ContinuousAndCategoricalArray):
      dof = sum(x.shape[-1] for x in jax.tree_util.tree_leaves(data.features))
    else:
      dof = len(specs)
    if data.label_is_missing is None:
      num_obs = data.labels.shape[0]
    else:
      num_obs = jnp.sum(~data.label_is_missing)
    # TODO: Discount the infeasible points. The 0.1 and 0.9 split
    # is associated with weights to feasible and infeasbile trials.
    trust_level = (0.1 * num_obs + 0.9 * num_obs) / (
        dimension_factor * (dof + 1)
    )
    trust_radius = min_radius + (0.5 - min_radius) * trust_level

    if isinstance(data.features, types.ContinuousAndCategoricalArray):
      max_distance = [np.inf] * data.features.continuous.shape[-1]
    else:
      max_distance = []
      for spec in specs:
        # Cap distances between one-hot encoded features so that they fall
        # within the trust region radius.
        if spec.type is converters.NumpyArraySpecType.ONEHOT_EMBEDDING:
          max_distance.extend([trust_radius] * spec.num_dimensions)
        else:
          max_distance.append(np.inf)
      if data.dimension_is_missing is not None:
        # These extra dimensions should be ignored.
        max_distance.extend(
            [0.0] * (data.dimension_is_missing.shape[-1] - len(max_distance))  # pytype: disable=attribute-error
        )

    max_distances = np.array(max_distance)
    return TrustRegion(max_distances, data, trust_radius)

  def min_linf_distance(self, xs: types.Features) -> jax.Array:
    """l-inf norm distance to the closest trusted point.

    Caps distances between one-hot encoded features to the trust-region radius,
    so that the trust region cutoff does not discourage exploration of these
    features.

    Args:
      xs: (M, D) array where each element is in [0, 1].

    Returns:
      (M,) array of floating numbers, L-infinity distances to the nearest
      trusted point.
    """
    trusted = self.data.features
    if isinstance(trusted, types.ContinuousAndCategoricalArray):
      trusted = trusted.continuous
    if isinstance(xs, types.ContinuousAndCategoricalArray):
      xs = xs.continuous
    if self.data.dimension_is_missing is not None:
      # Mask out padded dimensions
      trusted = jnp.where(
          self.data.dimension_is_missing, 0.0, self.data.features
      )
      xs = jnp.where(self.data.dimension_is_missing, jnp.zeros_like(xs), xs)
    distances = jnp.abs(trusted - xs[..., jnp.newaxis, :])  # (M, N, D)
    if self.data.label_is_missing is not None:
      # Mask out padded features. We set these distances to infinite since
      # they should never be considered.
      distances = jnp.where(
          self.data.label_is_missing[..., jnp.newaxis], np.inf, distances
      )
    distances_bounded = jnp.minimum(distances, self.max_distances)
    linf_distance = jnp.max(distances_bounded, axis=-1)  # (M, N)
    return jnp.min(linf_distance, axis=-1)  # (M,)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Implementation of SQL Datastore."""
import collections
import threading
from typing import Callable, DefaultDict, Iterable, List, Optional
from absl import logging

import sqlalchemy as sqla

from vizier._src.service import custom_errors
from vizier._src.service import datastore
from vizier._src.service import key_value_pb2
from vizier._src.service import resources
from vizier._src.service import study_pb2
from vizier._src.service import vizier_oss_pb2
from vizier.service import pyvizier as vz

from google.longrunning import operations_pb2


# TODO: Consider using ORM API (when fixed) to reduce code length.
class SQLDataStore(datastore.DataStore):
  """SQL Datastore."""

  def __init__(self, engine):
    self._engine = engine
    self._connection = self._engine.connect()
    self._root_metadata = sqla.MetaData()
    self._owners_table = sqla.Table(
        'owners',
        self._root_metadata,
        sqla.Column('owner_name', sqla.String, primary_key=True),
    )
    self._studies_table = sqla.Table(
        'studies',
        self._root_metadata,
        sqla.Column('study_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('serialized_study', sqla.String),
    )
    self._trials_table = sqla.Table(
        'trials',
        self._root_metadata,
        sqla.Column('trial_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_trial', sqla.String),
    )
    self._suggestion_operations_table = sqla.Table(
        'suggestion_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('client_id', sqla.String),
        sqla.Column('operation_number', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    self._early_stopping_operations_table = sqla.Table(
        'early_stopping_operations',
        self._root_metadata,
        sqla.Column('operation_name', sqla.String, primary_key=True),
        sqla.Column('owner_id', sqla.String),
        sqla.Column('study_id', sqla.String),
        sqla.Column('trial_id', sqla.INTEGER),
        sqla.Column('serialized_op', sqla.String),
    )
    # This lock is meant to lock `execute()` calls for database types which
    # don't support multi-threading, like SQLite.
    self._lock = threading.Lock()
    self._root_metadata.create_all(self._engine)

  def create_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    owner_name = study_resource.owner_resource.name
    owner_query = self._owners_table.insert().values(owner_name=owner_name)
    study_query = self._studies_table.insert().values(
        study_name=study.name,
        owner_id=study_resource.owner_id,
        study_id=study_resource.study_id,
        serialized_study=study.SerializeToString(),
    )

    with self._lock:
      try:
        self._connection.execute(owner_query)
      except sqla.exc.IntegrityError:
        logging.info('Owner with name %s currently exists.', owner_name)
      try:
        self._connection.execute(study_query)
        return study_resource
      except sqla.exc.IntegrityError as integrity_error:
        raise custom_errors.AlreadyExistsError(
            'Study with name %s already exists.' % study.name
        ) from integrity_error

  def load_study(self, study_name: str) -> study_pb2.Study:
    query = sqla.select([self._studies_table])
    query = query.where(self._studies_table.c.study_name == study_name)

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find study name: %s' % study_name
      )
    return study_pb2.Study.FromString(row['serialized_study'])

  def update_study(self, study: study_pb2.Study) -> resources.StudyResource:
    study_resource = resources.StudyResource.from_name(study.name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study.name
        )
    ).select()
    update_query = (
        sqla.update(self._studies_table)
        .where(self._studies_table.c.study_name == study.name)
        .values(
            study_name=study.name,
            owner_id=study_resource.owner_id,
            study_id=study_resource.study_id,
            serialized_study=study.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study.name
        )
      self._connection.execute(update_query)
    return study_resource

  def delete_study(self, study_name: str) -> None:
    study_resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.select([self._studies_table])
    exists_query = exists_query.where(
        self._studies_table.c.study_name == study_name
    )
    exists_query = sqla.exists(exists_query).select()
    delete_study_query = self._studies_table.delete().where(
        self._studies_table.c.study_name == study_name
    )
    delete_trials_query = (
        self._trials_table.delete()
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study_name
        )
      self._connection.execute(delete_study_query)
      self._connection.execute(delete_trials_query)

  def list_studies(self, owner_name: str) -> List[study_pb2.Study]:
    owner_id = resources.OwnerResource.from_name(owner_name).owner_id
    exists_query = sqla.exists(
        sqla.select([self._owners_table]).where(
            self._owners_table.c.owner_name == owner_name
        )
    ).select()
    list_query = sqla.select([self._studies_table]).where(
        self._studies_table.c.owner_id == owner_id
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Owner name %s does not exist.' % owner_name
        )
      result = self._connection.execute(list_query).fetchall()

    return [
        study_pb2.Study.FromString(row['serialized_study']) for row in result
    ]

  def create_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    query = self._trials_table.insert().values(
        trial_name=trial.name,
        owner_id=trial_resource.owner_id,
        study_id=trial_resource.study_id,
        trial_id=trial_resource.trial_id,
        serialized_trial=trial.SerializeToString(),
    )

    with self._lock:
      try:
        self._connection.execute(query)
        return trial_resource
      except sqla.exc.IntegrityError as integrity_error:
        raise custom_errors.AlreadyExistsError(
            'Trial with name %s already exists.' % trial.name
        ) from integrity_error

  def get_trial(self, trial_name: str) -> study_pb2.Trial:
    query = sqla.select([self._trials_table])
    query = query.where(self._trials_table.c.trial_name == trial_name)

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find trial name: %s' % trial_name
      )
    return study_pb2.Trial.FromString(row['serialized_trial'])

  def update_trial(self, trial: study_pb2.Trial) -> resources.TrialResource:
    trial_resource = resources.TrialResource.from_name(trial.name)
    exists_query = sqla.exists(
        sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial.name
        )
    ).select()
    update_query = (
        sqla.update(self._trials_table)
        .where(self._trials_table.c.trial_name == trial.name)
        .values(
            trial_name=trial.name,
            owner_id=trial_resource.owner_id,
            study_id=trial_resource.study_id,
            trial_id=trial_resource.trial_id,
            serialized_trial=trial.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Trial %s does not exist.' % trial.name
        )
      self._connection.execute(update_query)

    return trial_resource

  def list_trials(self, study_name: str) -> List[study_pb2.Trial]:
    study_resource = resources.StudyResource.from_name(study_name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study_name
        )
    ).select()
    list_query = (
        sqla.select([self._trials_table])
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study name %s does not exist.' % study_name
        )
      result = self._connection.execute(list_query)

    return [
        study_pb2.Trial.FromString(row['serialized_trial']) for row in result
    ]

  def delete_trial(self, trial_name: str) -> None:
    exists_query = sqla.exists(
        sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial_name
        )
    ).select()
    delete_query = self._trials_table.delete().where(
        self._trials_table.c.trial_name == trial_name
    )
    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Trial %s does not exist.' % trial_name
        )
      self._connection.execute(delete_query)

  def max_trial_id(self, study_name: str) -> int:
    study_resource = resources.StudyResource.from_name(study_name)
    exists_query = sqla.exists(
        sqla.select([self._studies_table]).where(
            self._studies_table.c.study_name == study_name
        )
    ).select()
    trial_id_query = (
        sqla.select(
            [sqla.func.max(self._trials_table.c.trial_id, type_=sqla.INT)]
        )
        .where(self._trials_table.c.owner_id == study_resource.owner_id)
        .where(self._trials_table.c.study_id == study_resource.study_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Study %s does not exist.' % study_name
        )
      potential_trial_id = self._connection.execute(trial_id_query).fetchone()[
          0
      ]

    if potential_trial_id is None:
      return 0
    return potential_trial_id

  def create_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)
    query = self._suggestion_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        client_id=resource.client_id,
        operation_number=resource.operation_number,
        serialized_op=operation.SerializeToString(),
    )

    try:
      with self._lock:
        self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise custom_errors.AlreadyExistsError(
          'Suggest Op with name %s already exists.' % operation.name
      ) from integrity_error

  def get_suggestion_operation(
      self, operation_name: str
  ) -> operations_pb2.Operation:
    query = sqla.select([self._suggestion_operations_table]).where(
        self._suggestion_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find suggest op name: %s' % operation_name
      )
    return operations_pb2.Operation.FromString(row['serialized_op'])

  def update_suggestion_operation(
      self, operation: operations_pb2.Operation
  ) -> resources.SuggestionOperationResource:
    resource = resources.SuggestionOperationResource.from_name(operation.name)

    exists_query = sqla.exists(
        sqla.select([self._suggestion_operations_table]).where(
            self._suggestion_operations_table.c.operation_name == operation.name
        )
    ).select()
    update_query = (
        sqla.update(self._suggestion_operations_table)
        .where(
            self._suggestion_operations_table.c.operation_name == operation.name
        )
        .values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            study_id=resource.study_id,
            client_id=resource.client_id,
            operation_number=resource.operation_number,
            serialized_op=operation.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Suggest op %s does not exist.' % operation.name
        )
      self._connection.execute(update_query)
    return resource

  def list_suggestion_operations(
      self,
      study_name: str,
      client_id: str,
      filter_fn: Optional[Callable[[operations_pb2.Operation], bool]] = None,
  ) -> List[operations_pb2.Operation]:
    study_resource = resources.StudyResource.from_name(study_name)
    query = sqla.select([self._suggestion_operations_table])
    query = query.where(
        self._suggestion_operations_table.c.owner_id == study_resource.owner_id
    )
    query = query.where(
        self._suggestion_operations_table.c.study_id == study_resource.study_id
    )
    query = query.where(
        self._suggestion_operations_table.c.client_id == client_id
    )

    exists_query = sqla.exists(query).select()
    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Could not find (study_name, client_id):',
            (study_resource.name, client_id),
        )
      result = self._connection.execute(query)

    all_ops = [
        operations_pb2.Operation.FromString(row['serialized_op'])
        for row in result
    ]
    if filter_fn is not None:
      output_list = []
      for op in all_ops:
        if filter_fn(op):
          output_list.append(op)
      return output_list
    else:
      return all_ops

  def max_suggestion_operation_number(
      self, study_name: str, client_id: str
  ) -> int:
    resource = resources.StudyResource.from_name(study_name)

    exists_query = sqla.exists(
        sqla.select([self._suggestion_operations_table])
        .where(
            self._suggestion_operations_table.c.owner_id == resource.owner_id
        )
        .where(
            self._suggestion_operations_table.c.study_id == resource.study_id
        )
        .where(self._suggestion_operations_table.c.client_id == client_id)
    ).select()
    max_query = (
        sqla.select(
            [
                sqla.func.max(
                    self._suggestion_operations_table.c.operation_number,
                    type_=sqla.INT,
                )
            ]
        )
        .where(
            self._suggestion_operations_table.c.owner_id == resource.owner_id
        )
        .where(
            self._suggestion_operations_table.c.study_id == resource.study_id
        )
        .where(self._suggestion_operations_table.c.client_id == client_id)
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Could not find (study_name, client_id):', (study_name, client_id)
        )
      return self._connection.execute(max_query).fetchone()[0]

  def create_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    query = self._early_stopping_operations_table.insert().values(
        operation_name=operation.name,
        owner_id=resource.owner_id,
        study_id=resource.study_id,
        trial_id=resource.trial_id,
        serialized_op=operation.SerializeToString(),
    )

    try:
      with self._lock:
        self._connection.execute(query)
      return resource
    except sqla.exc.IntegrityError as integrity_error:
      raise custom_errors.AlreadyExistsError(
          'Early stopping op with name %s already exists.' % operation.name
      ) from integrity_error

  def get_early_stopping_operation(
      self, operation_name: str
  ) -> vizier_oss_pb2.EarlyStoppingOperation:
    query = sqla.select([self._early_stopping_operations_table]).where(
        self._early_stopping_operations_table.c.operation_name == operation_name
    )

    with self._lock:
      result = self._connection.execute(query)

    row = result.fetchone()
    if not row:
      raise custom_errors.NotFoundError(
          'Failed to find early stopping op name: %s' % operation_name
      )
    return vizier_oss_pb2.EarlyStoppingOperation.FromString(
        row['serialized_op']
    )

  def update_early_stopping_operation(
      self, operation: vizier_oss_pb2.EarlyStoppingOperation
  ) -> resources.EarlyStoppingOperationResource:
    resource = resources.EarlyStoppingOperationResource.from_name(
        operation.name
    )
    exists_query = sqla.exists(
        sqla.select([self._early_stopping_operations_table]).where(
            self._early_stopping_operations_table.c.operation_name
            == operation.name
        )
    ).select()
    update_query = (
        sqla.update(self._early_stopping_operations_table)
        .where(
            self._early_stopping_operations_table.c.operation_name
            == operation.name
        )
        .values(
            operation_name=operation.name,
            owner_id=resource.owner_id,
            study_id=resource.study_id,
            trial_id=resource.trial_id,
            serialized_op=operation.SerializeToString(),
        )
    )

    with self._lock:
      exists = self._connection.execute(exists_query).fetchone()[0]
      if not exists:
        raise custom_errors.NotFoundError(
            'Early stopping op %s does not exist.' % operation.name
        )
      self._connection.execute(update_query)
      return resource

  def update_metadata(
      self,
      study_name: str,
      study_metadata: Iterable[key_value_pb2.KeyValue],
      trial_metadata: Iterable[datastore.UnitMetadataUpdate],
  ) -> None:
    """Store the supplied metadata into the SQL database."""
    s_resource = resources.StudyResource.from_name(study_name)
    logging.debug('database.update_metadata s_resource= %s', s_resource)
    # Obtain original study.
    get_study_query = sqla.select([self._studies_table]).where(
        self._studies_table.c.study_name == study_name
    )

    with self._lock:
      study_result = self._connection.execute(get_study_query)
      row = study_result.fetchone()
      if not row:
        raise custom_errors.NotFoundError('No such study:', s_resource.name)
      original_study = study_pb2.Study.FromString(row['serialized_study'])

      # Store Study-related metadata into the database.
      vz.metadata_util.merge_study_metadata(
          original_study.study_spec, study_metadata
      )
      update_study_query = (
          sqla.update(self._studies_table)
          .where(self._studies_table.c.study_name == study_name)
          .values(serialized_study=original_study.SerializeToString())
      )
      self._connection.execute(update_study_query)

      # Split the trial-related metadata by Trial.
      split_metadata: DefaultDict[str, List[datastore.UnitMetadataUpdate]] = (
          collections.defaultdict(list)
      )
      for md in trial_metadata:
        split_metadata[md.trial_id].append(md)

      # Now, we update one Trial at a time:
      for trial_id, md_list in split_metadata.items():
        t_resource = s_resource.trial_resource(trial_id)

        # Obtain original trial.
        trial_name = t_resource.name
        original_trial_query = sqla.select([self._trials_table]).where(
            self._trials_table.c.trial_name == trial_name
        )
        trial_result = self._connection.execute(original_trial_query)
        row = trial_result.fetchone()
        if not row:
          raise custom_errors.NotFoundError('No such trial:', trial_name)
        original_trial = study_pb2.Trial.FromString(row['serialized_trial'])

        # Update Trial.
        vz.metadata_util.merge_trial_metadata(original_trial, md_list)
        update_trial_query = (
            sqla.update(self._trials_table)
            .where(self._trials_table.c.trial_name == trial_name)
            .values(serialized_trial=original_trial.SerializeToString())
        )
        self._connection.execute(update_trial_query)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for eagle strategy."""

from absl.testing import parameterized
import jax
from jax import numpy as jnp
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import eagle_param_handler
from vizier._src.algorithms.optimizers import eagle_strategy
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier.pyvizier import converters
from vizier.pyvizier.converters import padding

from absl.testing import absltest


def _create_features_simple(
    features, rewards, features_batch, rewards_batch, config, n_features
):
  """A version of `_create_features` that materializes large intermediates."""
  features_diffs = features - features_batch[:, jnp.newaxis, :]
  dists = jnp.sum(jnp.square(features_diffs), axis=-1)
  directions = rewards - rewards_batch[:, jnp.newaxis]
  scaled_directions = jnp.where(
      directions >= 0.0, config.gravity, -config.negative_gravity
  )

  # Normalize the distance by the number of features.
  force = jnp.exp(-config.visibility * dists / n_features * 10.0)
  scaled_force = scaled_directions * force
  # Handle removed fireflies without updated rewards.
  finite_ind = jnp.isfinite(rewards).astype(scaled_force.dtype)

  # Ignore fireflies that were removed from the pool.
  scaled_force = scaled_force * finite_ind

  # Separate forces to pull and push so to normalize them separately.
  scaled_pulls = jnp.maximum(scaled_force, 0.0)
  scaled_push = jnp.minimum(scaled_force, 0.0)
  features_changes = jnp.sum(
      features_diffs * (scaled_pulls + scaled_push)[..., jnp.newaxis], axis=1
  )
  return features_batch + features_changes


class VectorizedEagleStrategyContinuousTest(parameterized.TestCase):

  def setUp(self):
    super(VectorizedEagleStrategyContinuousTest, self).setUp()
    self.config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    self.eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=self.config
    )(converter=converter, suggestion_batch_size=2)

  def test_create_features(self):
    features = jnp.array([[1, 2], [3, 4], [7, 7], [8, 8]])
    rewards = jnp.array([2, 3, 4, 1])
    seed = jax.random.PRNGKey(0)
    features_batch = features[: self.eagle.batch_size]
    rewards_batch = rewards[: self.eagle.batch_size]
    self.assertEqual(
        self.eagle._create_features(
            features,
            rewards,
            features_batch,
            rewards_batch,
            seed=seed,
        ).shape,
        (2, 2),
    )

    expected = _create_features_simple(
        features,
        rewards,
        features_batch,
        rewards_batch,
        self.config.replace(
            mutate_normalization_type=(
                eagle_strategy.MutateNormalizationType.UNNORMALIZED
            )
        ),
        self.eagle.n_feature_dimensions,
    )
    actual = self.eagle._create_features(
        features,
        rewards,
        features_batch,
        rewards_batch,
        seed=seed,
    )
    np.testing.assert_array_equal(expected, actual)

  def test_create_random_perturbations(self):
    seed = jax.random.PRNGKey(0)
    perturbations_batch = jnp.ones(self.eagle.batch_size)
    perturbations = self.eagle._create_random_perturbations(
        perturbations_batch, seed=seed
    )
    self.assertEqual(perturbations.shape, (2, 2))

  def test_update_pool_features_and_rewards(self):
    features = jnp.array([[1, 2], [3, 4], [7, 7], [8, 8]], dtype=jnp.float64)
    rewards = jnp.array([2, 3, 4, 1], dtype=jnp.float64)
    perturbations = jnp.array([1, 1, 1, 1], dtype=jnp.float64)

    batch_features = jnp.array([[9, 9], [10, 10]], dtype=jnp.float64)
    batch_rewards = jnp.array([5, 0.5], dtype=jnp.float64)

    new_features, new_rewards, new_perturbations = (
        self.eagle._update_pool_features_and_rewards(
            batch_features,
            batch_rewards,
            features[: self.eagle.batch_size],
            rewards[: self.eagle.batch_size],
            perturbations[: self.eagle.batch_size],
        )
    )
    np.testing.assert_array_equal(
        new_features,
        np.array([[9, 9], [3, 4]], dtype=np.float64),
        err_msg='Features are not equal.',
    )

    np.testing.assert_array_equal(
        new_rewards,
        np.array([5, 3], dtype=np.float64),
        err_msg='rewards are not equal.',
    )

    pc = self.config.penalize_factor
    np.testing.assert_array_almost_equal(
        new_perturbations,
        np.array([1, pc], dtype=np.float64),
        err_msg='Perturbations are not equal.',
    )

  def test_update_best_reward(self):
    # Test replacing the best reward.
    features = jnp.array([[1, 2], [3, 4], [7, 7], [8, 8]], dtype=jnp.float64)
    rewards = jnp.array([2, 3, 4, 1], dtype=jnp.float64)
    state = eagle_strategy.VectorizedEagleStrategyState(
        iterations=jnp.array(0),
        features=features,
        rewards=rewards,
        best_reward=jnp.max(rewards),
        perturbations=jnp.ones_like(rewards),
    )
    batch_features = jnp.array([[9, 9], [10, 10]], dtype=jnp.float64)
    batch_rewards = jnp.array([5, 0.5], dtype=jnp.float64)
    seed = jax.random.PRNGKey(0)
    new_state = self.eagle.update(
        state, batch_features, batch_rewards, seed=seed
    )
    self.assertEqual(new_state.best_reward, 5.0)
    # Test not replacing the best reward.
    batch_rewards = jnp.array([2, 4], dtype=jnp.float64)
    new_new_state = self.eagle.update(
        new_state, batch_features, batch_rewards, seed=seed
    )
    self.assertEqual(new_new_state.best_reward, 5.0)

  @parameterized.parameters(
      {'batch_size': 5, 'expected_batch_size': 5, 'max_pool_size': 50},
      {'batch_size': None, 'expected_batch_size': 50, 'max_pool_size': 50},
      {'batch_size': 5, 'expected_batch_size': 5, 'max_pool_size': 10},
      {'batch_size': None, 'expected_batch_size': 10, 'max_pool_size': 10},
  )
  def test_batch_size_and_pool_size(
      self, batch_size, expected_batch_size, max_pool_size
  ):
    problem = vz.ProblemStatement()
    root = problem.search_space.root
    for i in range(100):
      root.add_float_param(f'x{i}', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    config = eagle_strategy.EagleStrategyConfig(max_pool_size=max_pool_size)
    eagle = eagle_strategy.VectorizedEagleStrategyFactory(eagle_config=config)(
        converter=converter, suggestion_batch_size=batch_size
    )
    self.assertEqual(eagle.pool_size, max_pool_size)
    self.assertEqual(eagle.batch_size, expected_batch_size)

  def test_trim_pool(self):
    pc = self.config.perturbation
    features_batch = jnp.array([[1, 2], [3, 4]], dtype=jnp.float64)
    rewards_batch = jnp.array([2, 3], dtype=jnp.float64)
    perturbations = jnp.array([pc, 0], dtype=jnp.float64)
    seed = jax.random.PRNGKey(0)
    new_features, new_rewards, new_perturbations = self.eagle._trim_pool(
        features_batch,
        rewards_batch,
        perturbations,
        best_reward=jnp.array(4.0),
        seed=seed,
    )

    np.testing.assert_array_almost_equal(
        new_features[0],
        features_batch[0],
        err_msg='Features are not equal.',
    )
    self.assertTrue(
        all(np.not_equal(new_features[1], features_batch[1])),
        msg='Features are not equal.',
    )

    np.testing.assert_array_equal(
        new_rewards,
        np.array([2, -np.inf], dtype=np.float64),
        err_msg='rewards are not equal.',
    )
    # The best firefly is never removed.
    np.testing.assert_array_almost_equal(
        new_perturbations,
        np.array([pc, pc], dtype=np.float64),
        err_msg='Perturbations are not equal.',
    )

  def test_create_strategy_from_factory(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    eagle = eagle_factory(converter)
    self.assertEqual(eagle.n_feature_dimensions, 3)

  def test_optimize_with_eagle(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        converter
    )
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    optimizer(score_fn=lambda x: -jnp.sum(x, 1), count=1)

  def test_optimize_with_eagle_padding(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    converter = converters.PaddedTrialToArrayConverter.from_study_config(
        problem,
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.POWERS_OF_2,
            num_features=padding.PaddingType.POWERS_OF_2,
        ),
    )
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    converter = converters.TrialToArrayConverter.from_study_config(problem)
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        converter
    )
    optimizer(score_fn=lambda x: -np.sum(x, 1), count=1)


class EagleParamHandlerTest(parameterized.TestCase):

  def setUp(self):
    super(EagleParamHandlerTest, self).setUp()
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_categorical_param('c1', ['a', 'b'])
    root.add_float_param('f1', 0.0, 5.0)
    root.add_categorical_param('c2', ['a', 'b', 'c'])
    root.add_discrete_param('d1', [2.0, 3.0, 5.0, 11.0])
    converter = converters.TrialToArrayConverter.from_study_config(
        problem, max_discrete_indices=0, pad_oovs=True
    )
    self.config = eagle_strategy.EagleStrategyConfig()
    self.param_handler = eagle_param_handler.EagleParamHandler.build(
        converter=converter,
        categorical_perturbation_factor=self.config.categorical_perturbation_factor,
        pure_categorical_perturbation_factor=self.config.pure_categorical_perturbation_factor,
    )

  def test_init(self):
    self.assertEqual(self.param_handler.n_feature_dimensions, 9)
    self.assertLen(
        self.param_handler.perturbation_factors,
        self.param_handler.n_feature_dimensions,
    )
    self.assertEqual(self.param_handler.n_categorical, 2)

  def test_categorical_params_mask(self):
    expected_categorical_params_mask = np.array(
        [[1, 1, 1, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 1, 1, 1, 1, 0]]
    )
    np.testing.assert_array_equal(
        self.param_handler._categorical_params_mask,
        expected_categorical_params_mask,
    )

  def test_categorical_mask(self):
    expected_categorical_mask = np.array([1, 1, 1, 0, 1, 1, 1, 1, 0])
    np.testing.assert_array_equal(
        self.param_handler._categorical_mask, expected_categorical_mask
    )

  def test_tiebreak_mask(self):
    eps = self.param_handler._epsilon
    expected_tiebreak_mask = np.array(
        [-eps * (i + 1) for i in range(9)], dtype=float
    )
    np.testing.assert_array_almost_equal(
        self.param_handler._tiebreak_array, expected_tiebreak_mask
    )

  def test_categorical_oov_mask(self):
    expected_oov_mask = np.array([1, 1, 0, 1, 1, 1, 1, 0, 1], dtype=float)
    np.testing.assert_array_equal(
        self.param_handler._oov_mask, expected_oov_mask
    )

  def test_perturbation_factors(self):
    cp = self.config.categorical_perturbation_factor
    expected_perturbation_factors = np.array(
        [cp, cp, cp, 1, cp, cp, cp, cp, 1], dtype=float
    )
    np.testing.assert_array_equal(
        self.param_handler.perturbation_factors, expected_perturbation_factors
    )

  def test_sample_categorical_features(self):
    # features shouldn't have values in oov_mask, and have the structure of:
    # [c1,c1,c1,f1,c2,c2,c2,c2,d1]
    features = jnp.array([
        [2.0, 0.0, 0.0, 1.5, 0.0, 0.0, 0.1, 0.0, 9.0],
        [3.0, 0.0, 0.0, 3.5, 5.0, 0.0, 0.0, 0.0, 8.0],
    ])
    expected_sampled_features = np.array([
        [1.0, 0.0, 0.0, 1.5, 0.0, 0.0, 1.0, 0.0, 9.0],
        [1.0, 0.0, 0.0, 3.5, 1.0, 0.0, 0.0, 0.0, 8.0],
    ])
    sampled_features = self.param_handler.sample_categorical(
        features, seed=jax.random.PRNGKey(0)
    )
    np.testing.assert_array_equal(sampled_features, expected_sampled_features)

  def test_prior_trials(self):
    config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4, prior_trials_pool_pct=1.0
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)

    prior_features = jnp.array([[1, -1], [2, 1], [3, 2], [4, 5]])
    prior_rewards = jnp.array([1, 2, 3, 4])
    eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=config,
    )(converter=converter, suggestion_batch_size=2)
    init_state = eagle.init_state(
        jax.random.PRNGKey(0), prior_features, prior_rewards
    )
    np.testing.assert_array_equal(
        init_state.features, jnp.flip(prior_features, axis=0)
    )

  @parameterized.parameters(2, 10)
  def test_prior_trials_with_too_few_or_many_trials(self, n_prior_trials):
    # Test proper populating of the pool in case the number of prior trials is
    # greater or lesser than the pool size.
    config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4, prior_trials_pool_pct=1.0
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    converter = converters.TrialToArrayConverter.from_study_config(problem)

    prior_features = np.random.randn(n_prior_trials, 2)
    prior_rewards = np.random.randn(n_prior_trials)

    eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=config,
    )(converter=converter, suggestion_batch_size=2)
    init_state = eagle.init_state(
        jax.random.PRNGKey(0), prior_features, prior_rewards
    )
    self.assertEqual(init_state.features.shape, (config.pool_size, 2))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""TunerPolicy."""

from typing import Optional, Sequence

from absl import logging
import attr
import pyglove as pg
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from vizier._src.pyglove import core


@attr.define
class TunerPolicy(pythia.Policy):
  """Pythia policy for custom multi-trial tuner algorithm.

  Note that study_config should be used if the user needs the Trials to
  faithfully use the parameter names from the original StudyConfig.
  """

  supporter: pythia.PolicySupporter = attr.field()
  _converter: converters.VizierConverter = attr.field()
  _algorithm: pg.geno.DNAGenerator = attr.field()
  _incorporated_trial_ids: set[int] = attr.field(factory=set)
  _early_stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = attr.field(
      default=None
  )

  @property
  def algorithm(self) -> pg.geno.DNAGenerator:
    return self._algorithm

  @property
  def early_stopping_policy(self) -> Optional[pg.tuning.EarlyStoppingPolicy]:
    return self._early_stopping_policy

  @property
  def _metric_names(self) -> Sequence[str]:
    return self._converter.metrics_to_optimize

  def update(self, tuner_trial: pg.tuning.Trial) -> bool:
    """Update a single tuner Trial.

    Args:
      tuner_trial: If the trial id was previously seen, update is no-op.

    Returns:
      True if the trial was added.
    """
    if tuner_trial.id in self._incorporated_trial_ids:
      return False
    logging.info(
        'Updating TunerTrial %s to algorithm: %s', tuner_trial, self._algorithm
    )
    reward = tuner_trial.get_reward_for_feedback(self._metric_names)
    if reward is not None:
      self._algorithm.feedback(tuner_trial.dna, reward)
      self._incorporated_trial_ids.add(tuner_trial.id)
      return True
    return False

  def suggest(self, request: pythia.SuggestRequest) -> pythia.SuggestDecision:
    logging.info('Tuner policy get new suggestions')
    completed_trials = self.supporter.GetTrials(
        status_matches=vz.TrialStatus.COMPLETED
    )

    mu = vz.MetadataDelta()
    n_trials_updated = 0
    for vizier_trial in completed_trials:
      tuner_trial = core.VizierTrial(self._converter, vizier_trial)
      metadata = dict(tuner_trial.dna.metadata)  # make a deep copy.
      if self.update(tuner_trial):
        n_trials_updated += 1
        # Serialize DNA metadata to Vizier trial metadata upon change.
        if pg.ne(tuner_trial.dna.metadata, metadata):
          mu.assign(
              constants.METADATA_NAMESPACE,
              constants.TRIAL_METADATA_KEY_DNA_METADATA,
              pg.to_json_str(tuner_trial.dna.metadata),
              trial=vizier_trial,
          )

    new_trials = []
    count = request.count or 1

    logging.info(
        (
            'The algorithm now reflects %s new trials.'
            'The algorithm has seen %s trials total and '
            'there are %s total completed trials in the study '
            'Now generating %s new suggestions.'
        ),
        n_trials_updated,
        len(self._incorporated_trial_ids),
        len(completed_trials),
        count,
    )
    for _ in range(request.count or 1):
      dna = self._algorithm.propose()
      if dna.spec is None:
        dna.use_spec(self._converter.dna_spec)
      trial = self._converter.to_trial(dna, fallback='return_dummy')
      new_trials.append(trial)
      logging.info(
          'algorithm : %s proposed DNA: %s which is translated to trial: %s',
          self._algorithm,
          repr(dna),
          trial,
      )

    return pythia.SuggestDecision(new_trials, mu)

  def early_stop(
      self, request: pythia.EarlyStopRequest
  ) -> pythia.EarlyStopDecisions:
    # TODO: Support early stopping in the future.
    return pythia.EarlyStopDecisions()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Pythia Eagle Strategy Designer.

Implements a variation of Eagle Strategy without the Levy random walk, aka
Firefly Algorithm (FA).

Reference: Yang XS. (2009) Firefly Algorithms for Multimodal Optimization.
In: Stochastic Algorithms: Foundations and Applications (SAGA) 2009.
DOI: https://doi.org/10.1007/978-3-642-04944-6_14


Firefly Algorithm Summary
=========================
FA is a genetic algorithm that maintains a pool of fireflies. Each
firefly emits a light whose intensity is non-decreasing in (or simply equal
to) the objective value. Each iteration, a firefly chases after a brighter
firefly, but the brightness it perceives decreases in distance. This allows
multiple "clusters" to form, as opposed to all fireflies collapsing to a
single point. Not included in the original algorithm, we added "repulsion" which
in addition to the "attraction" forces, meaning fireflies move towards the
bright spots as well as away from the dark spots. We also support non-decimal
parameter types (categorical, discrete, integer), and treat them uniquely when
computing distance, adding pertrubation, and mutating fireflies.

For more details, see the linked paper.

OSS Vizier Implementation Summary
=================================
We maintain a pool of fireflies. Each firefly stores the best trial it created.
During 'Suggest' each firefly on its turn is used to suggest a new trial. The
best trial parameters are used to compute distances and to generate new
suggested parameters by calling the '_mutate' and '_perturb' methods. During
'Update' the trial results comeback and we create a new firefly if needed or
update/remove the associated firefly in the pool. To facilitate this
association, during 'Suggest', the newly created suggested trial stores in
its metadata the id of its parent firefly ('parent_fly_id'). Throughout the
study lifetime, the same firefly could be associated with multiple trials
created from it.

In our stateful implementation, across 'Suggest' calls, we persist the pool of
fireflies and update it using previously unseen COMPLETED trials. We also
persist the last firefly used to generate a suggestion so to not use
sequentially the same firefly even if there are ACTIVE trials. Lastly we persist
the maximum value of firefly id created, to ensure that each firefly has its own
unique id.
"""

import json
import logging
import time
from typing import Optional, Sequence

import attr
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy_utils
from vizier._src.algorithms.designers.eagle_strategy import serialization
from vizier._src.algorithms.random import random_sample
from vizier.interfaces import serializable
from vizier.pyvizier import converters


EagleStrategyUtils = eagle_strategy_utils.EagleStrategyUtils
FireflyAlgorithmConfig = eagle_strategy_utils.FireflyAlgorithmConfig
Firefly = eagle_strategy_utils.Firefly
FireflyPool = eagle_strategy_utils.FireflyPool


class EagleStrategyDesigner(vza.PartiallySerializableDesigner):
  """The eagle strategy partially serializable designer."""

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      *,
      config: Optional[FireflyAlgorithmConfig] = None,
      seed: Optional[int] = None,
      verbose: bool = True,
  ):
    """Initializes the Eagle Strategy desiger.

    Args:
      problem_statement: A problem description including the search space.
      config: The Firefly algorithm hyperparameters.
      seed: A seed to deterministically generate samples from random variables.
      verbose: Whether to display

    Raises:
      Exception: if the problem statement includes condional search space,
        mutli-objectives or safety metrics.
    """
    # Problem statement validation.
    if problem_statement.search_space.is_conditional:
      raise ValueError(
          "Eagle Strategy designer doesn't support conditional parameters."
      )
    if not problem_statement.is_single_objective:
      raise ValueError(
          "Eagle Strategy designer doesn't support multi-objectives."
      )
    if problem_statement.is_safety_metric:
      raise ValueError(
          "Eagle Strategy designer doesn't support safety metrics."
      )

    if seed is None:
      # When a key is not provided it will be set based on the current time to
      # ensure non-repeated behavior.
      seed = int(time.time())
      logging.info(
          ('A seed was not provided to Eagle Strategy designer constructor. '
           'Setting the seed to %s'),
          str(seed),
      )
    self._scaler = converters.ProblemAndTrialsScaler(problem_statement)
    self._problem = self._scaler.problem_statement
    self._rng = np.random.default_rng(seed=seed)
    self._config = config or FireflyAlgorithmConfig()
    self._utils = EagleStrategyUtils(self._problem, self._config, self._rng)
    self._firefly_pool = FireflyPool(
        utils=self._utils, capacity=self._utils.compute_pool_capacity())

    logging.info(
        (
            'Eagle Strategy designer initialized. Pool capacity: %s. '
            'Eagle config:\n%s\nProblem statement:\n%s'
        ),
        self._utils.compute_pool_capacity(),
        json.dumps(attr.asdict(self._config), indent=2),
        self._problem,
    )

  def dump(self) -> vz.Metadata:
    """Dumps the current state of the algorithm.

    Returns:
      Metadata with the current pool serialized.
    """
    metadata = vz.Metadata()
    metadata.ns('eagle')['rng'] = serialization.serialize_rng(self._rng)
    metadata.ns('eagle')['firefly_pool'] = (
        serialization.partially_serialize_firefly_pool(self._firefly_pool))
    metadata.ns('eagle')['serialization_version'] = 'v1'
    metadata.ns('eagle')['dump_timestamp'] = str(time.time())
    logging.info('Dump metadata:\n%s', metadata.ns('eagle'))
    return metadata

  def load(self, metadata: vz.Metadata) -> None:
    """Loads the current state of the algorithm run.

    The function is used as part of the `_restore_designer` method in
    PartiallySerializableDesignerPolicy to restore the state of the designer
    after it was already initialized with the problem statement.

    1. Set EagleStrategy, EagleStrategyUtils with the recovered random generator
    2. Recover the FireflyPool and populate it with the Fireflies.

    Args:
      metadata: Metadata
    """
    if metadata.ns('eagle').get('serialization_version', default=None) is None:
      # First time the designer is called, so the namespace doesn't exist yet.
      logging.info('Eagle designer was called for the first time. No state was'
                   ' recovered.')
    else:
      try:
        logging.info('Load metadata:\n%s', metadata.ns('eagle'))
        self._rng = serialization.restore_rng(metadata.ns('eagle')['rng'])
      except Exception as e:
        raise serializable.FatalDecodeError(
            "Couldn't load random generator from metadata.") from e
      self._utils.rng = self._rng
      logging.info('Restored rng:\n%s', self._rng)
      try:
        firefly_pool = metadata.ns('eagle')['firefly_pool']
        self._firefly_pool = serialization.restore_firefly_pool(
            self._utils, firefly_pool)
        logging.info('Restored firefly pool:\n%s', self._firefly_pool)
      except Exception as e:
        raise serializable.HarmlessDecodeError(
            "Couldn't load firefly pool from metadata.") from e
      logging.info(
          ('Eagle designer restored state from timestamp %s. Firefly pool'
           ' now contains %s fireflies.'),
          metadata.ns('eagle')['dump_timestamp'],
          self._firefly_pool.size,
      )

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggests trials."""
    scaled_suggestions = [self._suggest_one() for _ in range(count)]
    #  Unscale suggestion parameters to the original search space.
    return self._scaler.unmap(scaled_suggestions)

  def _suggest_one(self) -> vz.TrialSuggestion:
    """Generates a single suggestion based on the current pool of flies.

    In order to generate a trial suggestion, we find the next fly that should be
    moved and create a copy of it. Then we mutate and perturb its trial
    parameters inplace and assign them to the suggested trial.

    Returns:
      The suggested trial
      The parent fly Id
    """
    suggested_trial = vz.TrialSuggestion()
    if self._firefly_pool.size < self._firefly_pool.capacity:
      # Pool is underpopulated. Generate a random trial parameters.
      # (b/243518714): Use random policy/designer to generate parameters.
      suggested_parameters = random_sample.sample_parameters(
          self._rng, self._problem.search_space
      )
      # Create a new parent fly id and assign it to the trial, this will be
      # used during Update to match the trial to its parent fly in the pool.
      parent_fly_id = self._firefly_pool.generate_new_fly_id()
      logging.info('Pool is underpopulated. Generated random trial parameters.')
    else:
      # The pool is full. Use a copy of the next fly in line to be moved.
      moving_fly = self._firefly_pool.get_next_moving_fly_copy()
      self._mutate_fly(moving_fly)
      self._perturb_fly(moving_fly)
      suggested_parameters = moving_fly.trial.parameters
      parent_fly_id = moving_fly.id_
      logging.info('Created a trial from parent fly ID: %s.', parent_fly_id)

    suggested_trial.parameters = suggested_parameters
    suggested_trial.metadata.ns('eagle')['parent_fly_id'] = str(parent_fly_id)
    logging.info(
        'Suggested trial: %s',
        self._utils.display_trial(suggested_trial.to_trial(-1)),
    )
    return suggested_trial

  def _mutate_fly(self, moving_fly: Firefly) -> None:
    """Mutates fly's trial parameters inplace.

    Apply pulls from the rest of the pool's flies on `moving_fly` to mutate
    its trial's parameters.

    Args:
      moving_fly: the fire from the pool to mutate its trial parameters.
    """
    # Access the moving fly's trial parameters to be modified.
    mutated_parameters = moving_fly.trial.parameters
    # Shuffle the ordering, so to apply the attracting and repelling forces
    # in random order every time. Creates a deep copy of the pool members.
    shuffled_flies = self._firefly_pool.get_shuffled_flies(self._rng)
    for other_fly in shuffled_flies:
      is_other_fly_better = self._utils.is_better_than(other_fly.trial,
                                                       moving_fly.trial)
      # Compute 'other_fly' pull weights by parameter type. In the paper
      # this is called "attractivness" and is denoted by beta(r).
      pull_weights = self._utils.compute_pull_weight_by_type(
          other_fly.trial.parameters, mutated_parameters, is_other_fly_better)
      # Apply the pulls from 'other_fly' on the moving fly's parameters.
      for param_config in self._problem.search_space.parameters:
        pull_weight = pull_weights[param_config.type]
        # Accentuate 'other_fly' pull using 'exploration_rate'.
        if pull_weight > 0.5:
          explore_pull_weight = (
              self._config.explore_rate * pull_weight
              + (1 - self._config.explore_rate) * 1.0
          )
        else:
          explore_pull_weight = self._config.explore_rate * pull_weight
        # Update the parameters using 'other_fly' and 'explore_pull_rate'.
        mutated_parameters[param_config.name] = (
            self._utils.combine_two_parameters(
                param_config,
                other_fly.trial.parameters,
                mutated_parameters,
                explore_pull_weight,
            ))

  def _perturb_fly(self, moving_fly: Firefly) -> None:
    """Perturbs the fly's trial parameters inplace.

    Apply random perturbation to the fly's parameter values based on the
    moving_fly's 'perturbation' value.

    Args:
      moving_fly: the fire from the pool to mutate its trial parameters.
    """
    suggested_parameters = moving_fly.trial.parameters
    perturbations = self._utils.create_perturbations(moving_fly.perturbation)
    for i, param_config in enumerate(self._problem.search_space.parameters):
      perturbed_value = self._utils.perturb_parameter(
          param_config,
          suggested_parameters[param_config.name].value,
          perturbations[i],
      )
      suggested_parameters[param_config.name] = perturbed_value

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the pool.

    Iterate over new completed trials and update the firefly pool. Every trial
    that was suggested from Eagle Strategy has a metadata data containing the
    parent fly id. For trials that were added to the study externally we assign
    a new parent fly id.

    Trials passed to 'update' are in the unscaled/original search space, and
    will be converted to the scaled search space, so that all other methods
    in the designer deal with scaled trial values.

    Arguments:
      completed: Trials in the original search space.
      all_active:
    """
    del all_active
    trials = self._scaler.map(completed.trials)
    for trial in trials:
      # Replaces trial metric name with a canonical metric name, which makes the
      # serialization and deserialization simpler.
      trial = self._utils.standardize_trial_metric_name(trial)
      if not trial.metadata.ns('eagle').get('parent_fly_id'):
        # Trial was not generated from Eagle Strategy. Set a new parent fly id.
        trial.metadata.ns('eagle')['parent_fly_id'] = str(
            self._firefly_pool.generate_new_fly_id())
        logging.info('Received a trial without parent_fly_id.')
      self._update_one(trial)

  def _update_one(self, trial: vz.Trial) -> None:
    """Update the pool using a single trial."""
    parent_fly_id = int(trial.metadata.ns('eagle').get('parent_fly_id'))
    parent_fly = self._firefly_pool.find_parent_fly(parent_fly_id)
    if parent_fly is None:
      logging.info('Parent fly ID: %s is not in pool.', parent_fly_id)
      if trial.infeasible:
        # Ignore infeasible trials without parent fly.
        logging.info('Got infeasible trial without a parent fly in the pool.')
      elif self._firefly_pool.size < self._firefly_pool.capacity:
        # Pool is below capacity. Create a new firefly or update existing one.
        logging.info(
            'Pool is below capacity (size=%s). Invoke create or update pool.',
            self._firefly_pool.size,
        )
        self._firefly_pool.create_or_update_fly(trial, parent_fly_id)
        return
      else:
        # Pool is at capacity. Try assigning a parent to utilize the trial info.
        parent_fly = self._assign_closest_parent(trial)

    if parent_fly is None:
      # Parent fly wasn't established. No need to continue.
      logging.info('Parent fly was not established.')
      return

    elif not trial.infeasible and self._utils.is_better_than(
        trial, parent_fly.trial):
      # If there's an improvement, we update the parent with the new trial.
      logging.info(
          'Good step:\nParent trial (val=%s): %s\nChild trial (val=%s): %s\n',
          self._utils.get_metric(parent_fly.trial),
          self._utils.display_trial(parent_fly.trial),
          self._utils.get_metric(trial),
          self._utils.display_trial(trial),
      )
      parent_fly.trial = trial
      parent_fly.generation += 1
    else:
      # If there's no improvement, we penalize the parent by decreasing its
      # exploration capability and potenitally remove it from the pool.
      logging.info(
          'Bad step:\nParent trial (val=%s): %s\nChild trial (val=%s): %s\n',
          self._utils.get_metric(parent_fly.trial),
          self._utils.display_trial(parent_fly.trial),
          self._utils.get_metric(trial),
          self._utils.display_trial(trial),
      )
      self._penalize_parent_fly(parent_fly, trial)

  def _assign_closest_parent(self, trial: vz.Trial) -> Optional[Firefly]:
    """Finds the closest parent fly and checks that the trial improves on it.

    Note that the trial's `parent_fly_id` won't exist in the pool when:

    1. The trial was randomly generated during 'Suggest', and therefore is not
    associated with any existing fly in the pool.

    2. The fly was removed from the pool but not all associated trials were
    processed yet.

    3. The trial was added to the study independently of Eagle Strategy
    suggestions.

    Args:
      trial:

    Returns:
      None or a fly from the pool that is closest to the trial.
    """
    closest_parent_fly = self._firefly_pool.find_closest_parent(trial)
    if self._utils.is_better_than(trial, closest_parent_fly.trial):
      # Only returns the closest fly if there's improvement. Otherwise, we don't
      # return it to not count it as a failure, as the closest parent is not
      # reponsible for it.
      return closest_parent_fly

  def _penalize_parent_fly(self, parent_fly: Firefly, trial: vz.Trial) -> None:
    """Penalizes a parent fly.

    The method is called on a fly after its generated trial didn't improve the
    objective function. The fly is penalized by decreasing its exploration
    capability and potentially removing it from the pool.

    Args:
      parent_fly: The parent fly with to be penalized.
      trial: The generated trial from the parent fly that didn't imporove.
    """
    if trial.parameters == parent_fly.trial.parameters:
      # If the new trial is identical to the parent trial, it means that the
      # fly is stuck, and so we increase its perturbation.
      parent_fly.perturbation = min(
          parent_fly.perturbation * 10, self._config.max_perturbation
      )
      logging.info(
          ('Penalize Parent Id: %s. Parameters are stuck. '
           'New perturbation factor: %s'),
          parent_fly.id_,
          parent_fly.perturbation,
      )
    else:
      # Otherwise, penalize the parent by decreasing its perturbation factor.
      parent_fly.perturbation *= 0.9
      logging.info(
          ('Penalize Parent Id: %s. Decrease perturbation factor. '
           'New perturbation factor: %s'),
          parent_fly.id_,
          parent_fly.perturbation,
      )
    if parent_fly.perturbation < self._config.perturbation_lower_bound:
      # If the perturbation factor is too low we attempt to eliminate the
      # unsuccessful parent fly from the pool.
      if self._firefly_pool.size == self._firefly_pool.capacity:
        # Only remove if the pool is at capacity. This is critical in studies
        # with few feasible/safe trials to retain the feasible trials.
        if not self._firefly_pool.is_best_fly(parent_fly):
          # Check that the fly is not the best one we have thus far.
          self._firefly_pool.remove_fly(parent_fly)
          logging.info('Removed fly ID: %s from pool.', parent_fly.id_)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""L-BFGS-B Strategy optimizer."""

from typing import Optional, Union

import attr
import jax
import jax.numpy as jnp
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import vectorized_base
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax.optimizers import optimizers
from vizier.pyvizier import converters


@attr.define(kw_only=True)
class LBFGSBOptimizer:
  """L-BFGS-B optimizer."""

  # Number of parallel runs of L-BFGS-B.
  random_restarts: int = attr.field(init=True, repr=False, default=25)
  # In parallel optimization (suggesting multiple candidates at once), this is
  # the number of candidates to consider at once. The score function is assumed
  # to score this many candidates together and output a scalar.
  num_parallel_candidates: Optional[int] = attr.field(default=None)

  def optimize(
      self,
      converter: converters.TrialToArrayConverter,
      score_fn: Union[
          vectorized_base.ParallelArrayScoreFunction,
          vectorized_base.ArrayScoreFunction,
      ],
      *,
      count: int = 1,
      seed: Optional[int] = None,
  ) -> list[vz.Trial]:
    """Optimize a continuous objective function using L-BFGS-B.

    Arguments:
      converter: Converter to map between trials and arrays.
      score_fn: Converts (batches of) features to scores.
      count: The number of best results to return.
      seed: Optional seed.

    Returns:
      The best trials found in the optimization.
    """
    if self.num_parallel_candidates and count > 1:
      # Note that we can't distinguish between 'BatchArrayScoreFunction' and
      # 'ParallelArrayScoreFunction' using 'isinstance' as they both have the
      # same function names.
      raise ValueError(
          "LBFGSBOptimizer doesn't support batch of batches (count > 1 is"
          " disallowed when num_parallel_candidates is set)."
      )
    optimize = optimizers.JaxoptLbfgsB(
        random_restarts=self.random_restarts, best_n=count
    )
    num_features = sum(spec.num_dimensions for spec in converter.output_specs)

    feature_shape = [num_features]
    if self.num_parallel_candidates is not None:
      feature_shape = [self.num_parallel_candidates, num_features]

    def setup(rng):
      return jax.random.uniform(rng, shape=feature_shape)

    rng = jax.random.PRNGKey(seed or 0)

    # Constraints are [0, 1].
    constraints = sp.Constraint(
        bounds=(np.zeros(feature_shape), np.ones(feature_shape))
    )

    # We'll assume score_fn is differentiable.
    # We also wrap the score function because the convention for optimizers is
    # to maximize, while L-BFGS-B always minimizes.
    def wrapped_score_fn(x):
      # Add a batch axis since `score_fn` is assumed to work on a batch of
      # trials.
      return -score_fn(x[jnp.newaxis, ...])[0], dict()

    new_features, _ = optimize(
        setup, wrapped_score_fn, rng, constraints=constraints
    )
    new_rewards = np.asarray(score_fn(new_features[jnp.newaxis, ...]))[0]

    if self.num_parallel_candidates is None:
      parameters = converter.to_parameters(new_features[jnp.newaxis, ...])
    else:
      parameters = converter.to_parameters(new_features)
    trials = []
    for i in range(len(parameters)):
      trial = vz.Trial(parameters=parameters[i])
      trial.complete(vz.Measurement({'acquisition': new_rewards}))
      trials.append(trial)
    return trials


@attr.define
class LBFGSBOptimizerFactory:
  """LBFGSB strategy optimizer factory."""

  def __call__(
      self, random_restarts: int, num_parallel_candidates: Optional[int] = None
  ) -> LBFGSBOptimizer:
    """Generates a new LBFGSBOptimizer object."""
    return LBFGSBOptimizer(
        num_parallel_candidates=num_parallel_candidates,
        random_restarts=random_restarts,
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utilities for GP Bandit."""

import functools
from typing import Optional

import jax
from jax import numpy as jnp
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import predictive_fns
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier.pyvizier.converters import feature_mapper


def precompute_cholesky(model, data, params):
  return model.apply(
      {'params': params},
      data.features,
      data.labels,
      method=model.precompute_predictive,
      mutable='predictive',
      observations_is_missing=data.label_is_missing,
  )


def optimize_acquisition(
    count: int,
    model: sp.StochasticProcessModel,
    acquisition_fn: acquisitions.AcquisitionFunction,
    optimizer: vb.VectorizedOptimizer,
    prior_features: types.Array,
    state: types.GPState,
    seed: jax.random.KeyArray,
    use_vmap: bool,
    trust_region: Optional[acquisitions.TrustRegion],
    # TODO: Remove this when Eagle optimizer takes
    # ContinuousAndCategoricalFeatures.
    mapper: Optional[feature_mapper.ContinuousCategoricalFeatureMapper],
) -> vb.VectorizedStrategyResults:
  """Optimize the acquisition function."""
  base_score_fn = functools.partial(
      predictive_fns.acquisition_on_array,
      model=model,
      acquisition_fn=acquisition_fn,
      state=state,
      trust_region=trust_region,
      use_vmap=use_vmap,
  )
  if mapper is not None:
    score_fn = lambda xs: base_score_fn(mapper.map(xs))
  else:
    score_fn = base_score_fn
  return optimizer(
      score_fn=score_fn, count=count, prior_features=prior_features, seed=seed
  )


def stochastic_process_model_loss_fn(
    params: types.ParameterDict,
    model: sp.StochasticProcessModel,
    data: types.StochasticProcessModelData,
    normalize: bool = False,
):
  """Loss function for a stochastic process model."""
  kwargs = {}
  if data.dimension_is_missing is not None:
    kwargs['dimension_is_missing'] = data.dimension_is_missing
  gp, mutables = model.apply(
      {'params': params},
      data.features,
      **kwargs,
      mutable=['losses', 'predictive'],
  )
  loss = -gp.log_prob(
      data.labels,
      is_missing=data.label_is_missing,
  ) + jax.tree_util.tree_reduce(jnp.add, mutables['losses'])
  if normalize:
    loss /= data.labels.shape[0]
  return loss, dict()


def stochastic_process_model_setup(
    key: jax.random.KeyArray,
    model: sp.StochasticProcessModel,
    data: types.StochasticProcessModelData,
):
  """Setup function for a stochastic process model."""
  kwargs = {}
  if data.dimension_is_missing is not None:
    kwargs['dimension_is_missing'] = data.dimension_is_missing
  return model.init(key, data.features, **kwargs)['params']



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for MaskFeatures."""

from absl.testing import parameterized
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax.models import mask_features

from absl.testing import absltest


class MaskFeaturesTest(parameterized.TestCase):

  @parameterized.parameters(2, 4, 6, 8)
  def testMaskingAndGradient(self, dims):
    half_dims = dims // 2
    xs = np.random.randn(10, half_dims)
    xs = np.concatenate([xs, np.full([10, half_dims], np.nan)], axis=-1)
    kernel = tfp.math.psd_kernels.MaternThreeHalves()
    kernel = mask_features.MaskFeatures(
        kernel,
        dimension_is_missing=np.array([False] * half_dims + [True] * half_dims),
    )
    f = lambda x: jnp.sum(kernel.matrix(x, x))
    value, grad = jax.value_and_grad(f)(xs)
    self.assertTrue(np.all(~np.isnan(value)))
    self.assertTrue(np.all(~np.isnan(grad)))


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Types library for vizier/_src/jax."""

from typing import Any, Generic, Iterable, Mapping, Optional, TypeVar, Union

from flax import struct
from flax.core import scope as flax_scope
import jax
from jax.typing import ArrayLike
import numpy as np
from vizier.pyvizier.converters import padding


# We define our own Array type since `jax.typing.Array` and `chex.Array` both
# include scalar types, which result in type errors when array
# methods/properties like `.shape` are accessed.
Array = Union[np.ndarray, jax.Array]

MaybePaddedArray = Union[Array, padding.PaddedArray]

ArrayTree = Union[ArrayLike, Iterable['ArrayTree'], Mapping[Any, 'ArrayTree']]

# An ArrayTree that allows None values.
ArrayTreeOptional = Union[
    ArrayLike,
    Iterable[Optional['ArrayTreeOptional']],
    Mapping[Any, Optional['ArrayTreeOptional']],
]
ParameterDict = flax_scope.Collection
ModelState = flax_scope.VariableDict


@struct.dataclass
class ContinuousAndCategoricalArray:
  continuous: Array
  categorical: Array


Features = TypeVar('Features', Array, ContinuousAndCategoricalArray)


@struct.dataclass
class StochasticProcessModelData(Generic[Features]):
  features: Features
  labels: Array
  label_is_missing: Optional[Array] = None
  dimension_is_missing: Optional[Features] = None


@struct.dataclass
class GPState:
  """State that changes at each iteration."""

  data: StochasticProcessModelData
  model_state: ModelState



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""PSD Kernel for masking out dimensions."""

import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.jax import types


class MaskFeatures(tfp.math.psd_kernels.FeatureTransformed):
  """Kernel for masking out features by setting them to zero.

  By masking out feature dimensions to zero (or to any other finite value), we
  ensure that when the Stochastic Process Model aware of this masking ingests
  these features (which may be NaN), we will get finite gradients. See
  https://github.com/tensorflow/probability/blob/main/discussion/where-nan.pdf
  for more details.
  """

  def __init__(
      self,
      kernel: tfp.math.psd_kernels.PositiveSemidefiniteKernel,
      dimension_is_missing: types.Array,
  ):
    self._kernel = kernel
    self._dimension_is_missing = dimension_is_missing

    super(MaskFeatures, self).__init__(
        kernel,
        transformation_fn=lambda x, *_: jnp.where(  # pylint:disable=g-long-lambda
            dimension_is_missing, jnp.zeros_like(x), x
        ),
    )

  @property
  def kernel(self):
    return self._kernel

  @property
  def dimension_is_missing(self):
    return self._dimension_is_missing

  def _batch_shape(self):
    return self.kernel.batch_shape

  def _batch_shape_tensor(self):
    return self.kernel.batch_shape_tensor()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for gp_bandit."""

from typing import Any, Optional
from unittest import mock

import chex
from jax.config import config
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_bandit
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.optimizers import eagle_strategy as es
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.algorithms.testing import test_runners
from vizier._src.jax import gp_bandit_utils
from vizier._src.jax import predictive_fns
from vizier._src.jax.optimizers import optimizers
from vizier.pyvizier import converters
from vizier.pyvizier.converters import padding
from vizier.testing import test_studies
from vizier.utils import profiler

from absl.testing import absltest
from absl.testing import parameterized


ensemble_ard_optimizer = optimizers.JaxoptLbfgsB(random_restarts=10, best_n=5)
noensemble_ard_optimizer = optimizers.JaxoptLbfgsB(random_restarts=5, best_n=1)


def _build_mock_continuous_array_specs(n):
  continuous_spec = mock.create_autospec(converters.NumpyArraySpec)
  continuous_spec.type = converters.NumpyArraySpecType.CONTINUOUS
  continuous_spec.num_dimensions = 1
  return [continuous_spec] * n


class GoogleGpBanditTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(iters=3, batch_size=5, num_seed_trials=5),
      dict(iters=5, batch_size=1, num_seed_trials=2),
      dict(ard_optimizer='ensemble'),
      dict(ard_optimizer='noensemble'),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          ard_optimizer='ensemble',
          use_categorical_kernel=True,
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          iters=5,
          batch_size=1,
          num_seed_trials=3,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          ard_optimizer='ensemble',
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          ard_optimizer='noensemble',
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          ard_optimizer='ensemble',
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
      ),
      dict(
          ard_optimizer='noensemble',
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
      ),
  )
  def test_on_flat_continuous_space(
      self,
      *,
      iters: int = 5,
      batch_size: int = 1,
      num_seed_trials: int = 1,
      ard_optimizer: Any = 'noensemble',
      padding_schedule: Optional[padding.PaddingSchedule] = None,
      use_categorical_kernel: bool = False,
      use_trust_region: bool = True,
  ):
    # We use string names so that test case names are readable. Convert them
    # to objects.
    if ard_optimizer == 'noensemble':
      ard_optimizer = noensemble_ard_optimizer
    elif ard_optimizer == 'ensemble':
      ard_optimizer = ensemble_ard_optimizer
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(),
        max_evaluations=10,
    )

    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=num_seed_trials,
        ard_optimizer=ard_optimizer,
        padding_schedule=padding_schedule,
        use_categorical_kernel=use_categorical_kernel,
        use_trust_region=use_trust_region,
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            batch_size=batch_size,
            verbose=1,
            validate_parameters=True,
        ).run_designer(designer),
        iters * batch_size,
    )
    # Test that latency tracking works.
    self.assertIn('VizierGPBandit.suggest', profiler.Storage().runtimes())
    runtimes = profiler.Storage().runtimes().get('VizierGPBandit.suggest')
    self.assertGreater(runtimes[0].total_seconds(), 0)

    quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        problem.search_space
    )
    predict_trials = quasi_random_sampler.suggest(count=7)
    prediction = designer.predict(predict_trials)
    self.assertLen(prediction.mean, 7)
    self.assertLen(prediction.stddev, 7)
    self.assertFalse(np.isnan(prediction.mean).any())
    self.assertFalse(np.isnan(prediction.stddev).any())

  @parameterized.parameters(
      dict(iters=3, batch_size=5, num_seed_trials=5),
      dict(iters=5, batch_size=1, num_seed_trials=2),
      dict(iters=5, batch_size=1, num_seed_trials=1),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.NONE,
              num_features=padding.PaddingType.NONE,
          ),
      ),
      dict(
          iters=3,
          batch_size=5,
          num_seed_trials=5,
          use_categorical_kernel=True,
      ),
  )
  def test_on_flat_mixed_space(
      self,
      iters: int,
      batch_size: int,
      num_seed_trials: int,
      padding_schedule: Optional[padding.PaddingSchedule] = None,
      use_categorical_kernel: bool = False,
      use_trust_region: bool = True,
  ):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(), max_evaluations=10
    )

    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=num_seed_trials,
        padding_schedule=padding_schedule,
        use_categorical_kernel=use_categorical_kernel,
        use_trust_region=use_trust_region,
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=iters,
            batch_size=batch_size,
            verbose=1,
            validate_parameters=True,
        ).run_designer(designer),
        iters * batch_size,
    )

    quasi_random_sampler = quasi_random.QuasiRandomDesigner(
        problem.search_space
    )
    predict_trials = quasi_random_sampler.suggest(count=7)
    prediction = designer.predict(predict_trials)
    self.assertLen(prediction.mean, 7)
    self.assertLen(prediction.stddev, 7)
    self.assertFalse(np.isnan(prediction.mean).any())
    self.assertFalse(np.isnan(prediction.stddev).any())

  def test_prediction_accuracy(self):
    search_space = vz.SearchSpace()
    search_space.root.add_float_param('x0', -5.0, 5.0)
    problem = vz.ProblemStatement(
        search_space=search_space,
        metric_information=vz.MetricsConfig(
            metrics=[
                vz.MetricInformation(
                    'obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE
                ),
            ]
        ),
    )
    f = lambda x: -((x - 0.5) ** 2)

    suggestions = quasi_random.QuasiRandomDesigner(
        problem.search_space, seed=1
    ).suggest(100)

    obs_trials = []
    for idx, suggestion in enumerate(suggestions):
      trial = suggestion.to_trial(idx)
      x = suggestions[idx].parameters['x0'].value
      trial.complete(vz.Measurement(metrics={'obj': f(x)}))
      obs_trials.append(trial)

    ard_optimizer = optimizers.JaxoptLbfgsB(random_restarts=8, best_n=5)
    gp_designer = gp_bandit.VizierGPBandit(problem, ard_optimizer=ard_optimizer)
    gp_designer.update(vza.CompletedTrials(obs_trials), vza.ActiveTrials())
    pred_trial = vz.Trial({'x0': 0.0})
    pred = gp_designer.predict([pred_trial])
    self.assertLess(np.abs(pred.mean[0] - f(0.0)), 2e-2)

  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.gp_bandit_utils.'
      'stochastic_process_model_loss_fn',
      wraps=chex.assert_max_traces(
          gp_bandit_utils.stochastic_process_model_loss_fn, n=1
      ),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.gp_bandit_utils.'
      'stochastic_process_model_setup',
      wraps=chex.assert_max_traces(
          gp_bandit_utils.stochastic_process_model_setup, n=1
      ),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.predictive_fns.'
      'predict_on_array',
      wraps=chex.assert_max_traces(predictive_fns.predict_on_array, n=1),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.predictive_fns.'
      'acquisition_on_array',
      # `acquisition_on_array` is traced twice because it's called on
      # prior_features as well as predictive features.
      wraps=chex.assert_max_traces(predictive_fns.acquisition_on_array, n=2),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.predictive_fns.'
      'sample_on_array',
      wraps=chex.assert_max_traces(predictive_fns.sample_on_array, n=1),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.gp_bandit_utils.'
      'precompute_cholesky',
      wraps=chex.assert_max_traces(gp_bandit_utils.precompute_cholesky, n=1),
  )
  @mock.patch(
      'vizier._src.algorithms.designers.gp_bandit.gp_bandit_utils.'
      'optimize_acquisition',
      wraps=chex.assert_max_traces(gp_bandit_utils.optimize_acquisition, n=1),
  )
  def test_jit_once(self, *args):
    del args
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name='metric', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    vectorized_optimizer_factory = vb.VectorizedOptimizerFactory(
        strategy_factory=es.VectorizedEagleStrategyFactory(),
        max_evaluations=10,
    )
    padding_schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.MULTIPLES_OF_10,
        num_features=padding.PaddingType.POWERS_OF_2,
    )
    designer = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=3,
        ard_optimizer=noensemble_ard_optimizer,
        padding_schedule=padding_schedule,
    )

    # Padding schedule should avoid retracing every iteration.
    test_runners.RandomMetricsRunner(
        problem,
        iters=5,
        batch_size=1,
        verbose=1,
        validate_parameters=True,
    ).run_designer(designer)

    # Retracing should not occur when a new VizierGPBandit instance is created.
    designer2 = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=3,
        ard_optimizer=noensemble_ard_optimizer,
        padding_schedule=padding_schedule,
    )
    test_runners.RandomMetricsRunner(
        problem,
        iters=5,
        batch_size=1,
        verbose=1,
        validate_parameters=True,
    ).run_designer(designer2)

    # Without padding, functions will be retraced.
    no_padding_schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.NONE,
        num_features=padding.PaddingType.NONE,
    )
    designer_no_pad = gp_bandit.VizierGPBandit(
        problem=problem,
        acquisition_optimizer_factory=vectorized_optimizer_factory,
        num_seed_trials=3,
        ard_optimizer=noensemble_ard_optimizer,
        padding_schedule=no_padding_schedule,
    )
    with self.assertRaisesRegex(AssertionError, 'traced > (1|2) times'):
      test_runners.RandomMetricsRunner(
          problem,
          iters=5,
          batch_size=1,
          verbose=1,
          validate_parameters=True,
      ).run_designer(designer_no_pad)


if __name__ == '__main__':
  # Jax disables float64 computations by default and will silently convert
  # float64s to float32s. We must explicitly enable float64.
  config.update('jax_enable_x64', True)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import random

import numpy as np
from scipy import stats
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.policies import designer_policy
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class HaltonTest(absltest.TestCase):

  def test_generate_primes(self):
    primes = quasi_random._generate_primes(100)
    # For each prime, make sure it is not evenly divisible by any number less
    # than itself.
    for p in primes:
      for n in range(2, int(p**0.5) + 1):
        self.assertNotEqual(0, p % n)

  def test_get_scrambled_halton_element(self):
    num_dimensions = 4
    generator = quasi_random._HaltonSequence(
        num_dimensions=num_dimensions, skip_points=1000, scramble=True
    )
    sequence = [generator.get_next_list() for i in range(1000)]

    self.assertLess(max(max(sequence)), 1)
    self.assertGreater(min(min(sequence)), 0)
    self.assertLen(sequence, 1000)
    self.assertLen(sequence[0], 4)

    # Test uniformity of Halton value outputs.
    sequence = np.array(sequence)
    for dim in range(num_dimensions):
      _, p_value = stats.kstest(
          sequence[:, dim], stats.uniform(loc=0.0, scale=1.0).cdf
      )
      # p_value greater than 0.9 roughly means we're very certain it's uniform.
      self.assertGreater(p_value, 0.9)

  def test_deterministic(self):
    generator_1 = quasi_random._HaltonSequence(
        num_dimensions=4, skip_points=100, primes_override=[3, 5, 7, 11]
    )
    generator_2 = quasi_random._HaltonSequence(
        num_dimensions=4, skip_points=100, primes_override=[3, 5, 7, 11]
    )
    self.assertEqual(
        [generator_1.get_next_list() for i in range(100)],
        [generator_2.get_next_list() for i in range(100)],
    )
    self.assertEqual(
        [generator_1.get_next_list() for i in range(100)],
        [generator_2.get_next_list() for i in range(100)],
    )

  def test_unscrambled_sequence(self):
    generator = quasi_random._HaltonSequence(
        num_dimensions=1,
        skip_points=0,
        primes_override=[3],
        scramble=False,
        seed=0,
    )
    sequence = [generator.get_next_list()[0] for _ in range(7)]

    expected_sequence = [
        1 / 3,
        2 / 3,
        1 / 9,
        4 / 9,
        7 / 9,
        2 / 9,
        5 / 9,
    ]

    self.assertSequenceAlmostEqual(sequence, expected_sequence)


class QuasiRandomTest(absltest.TestCase):

  def test_on_flat_space(self):
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())
    designer = quasi_random.QuasiRandomDesigner(problem.search_space)
    self.assertLen(
        test_runners.run_with_random_metrics(
            designer, problem, iters=50, batch_size=5, validate_parameters=True
        ),
        250,
    )

  def test_dump_and_load(self):
    # Check metadata checkpointing.
    problem = vz.ProblemStatement(test_studies.flat_space_with_all_types())

    designer = quasi_random.QuasiRandomDesigner(problem.search_space)
    designer.suggest(random.randint(1, 50))
    metadata = designer.dump()

    designer2 = quasi_random.QuasiRandomDesigner(problem.search_space)
    designer2.load(metadata)

    num_suggestions = 10
    self.assertEqual(
        designer.suggest(num_suggestions), designer2.suggest(num_suggestions)
    )

  def test_distribution(self):
    # Make sure output distribution makes sense.
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer = quasi_random.QuasiRandomDesigner(problem.search_space)

    suggestions = designer.suggest(2000)
    float_points = [
        suggestion.parameters['float'].value for suggestion in suggestions
    ]

    # Test uniformity of end-to-end parameter values.
    # p_value greater than 0.9 roughly means we're very certain it's uniform.
    # Unfortunately KS-test doesn't work for discrete/categorical distributions.
    _, float_p_value = stats.kstest(
        float_points, stats.uniform(loc=0.0, scale=1.0).cdf
    )
    self.assertGreater(float_p_value, 0.9)

  def test_equal_seeds(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer_1 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_1 = designer_1.suggest(10)
    designer_2 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_2 = designer_2.suggest(10)
    self.assertEqual(suggestions_1, suggestions_2)

  def test_distinct_seeds(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    designer_1 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=0)
    suggestions_1 = designer_1.suggest(10)
    designer_2 = quasi_random.QuasiRandomDesigner(problem.search_space, seed=1)
    suggestions_2 = designer_2.suggest(10)
    self.assertNotEqual(suggestions_1, suggestions_2)

  def test_policy_wrapping(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('float', 0.0, 1.0)
    policy_supporter = pythia.InRamPolicySupporter(problem)
    policy = designer_policy.PartiallySerializableDesignerPolicy(
        problem,
        policy_supporter,
        quasi_random.QuasiRandomDesigner.from_problem,
    )

    # Make sure outputs are distinct.
    all_suggestions = []
    for _ in range(1000):
      request = pythia.SuggestRequest(
          study_descriptor=policy_supporter.study_descriptor(), count=1
      )
      decisions = policy.suggest(request)
      all_suggestions.extend(decisions.suggestions)

    distinct_suggestions = set(
        [
            tuple(suggestion.parameters.as_dict().values())
            for suggestion in all_suggestions
        ]
    )
    self.assertLen(distinct_suggestions, 1000)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for PaddedTrialToArrayConverter."""

from vizier import pyvizier
from vizier.pyvizier.converters import padded_trial_to_array_converter as pttac
from vizier.pyvizier.converters import padding

from absl.testing import absltest
from absl.testing import parameterized


Trial = pyvizier.Trial


class PaddedTrialToArrayConverterTest(parameterized.TestCase):

  @parameterized.parameters([
      dict(
          num_continuous=6,
          num_integer=1,
          categorical_size=5,
          num_trials=7,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.MULTIPLES_OF_10,
              num_features=padding.PaddingType.POWERS_OF_2,
          ),
          expected_num_trials=10,
          expected_dimension=16,
      ),
      dict(
          num_continuous=2,
          num_integer=4,
          categorical_size=3,
          num_trials=7,
          padding_schedule=padding.PaddingSchedule(
              num_trials=padding.PaddingType.POWERS_OF_2,
              num_features=padding.PaddingType.MULTIPLES_OF_10,
          ),
          expected_num_trials=8,
          expected_dimension=10,
      ),
  ])
  def test_padding(
      self,
      num_continuous,
      num_integer,
      categorical_size,
      num_trials,
      padding_schedule,
      expected_num_trials,
      expected_dimension,
  ):
    space = pyvizier.SearchSpace()
    root = space.root
    for i in range(num_continuous):
      root.add_float_param(f'double_{i}', 0.0, 100.0)

    for i in range(num_integer):
      root.add_int_param(f'integer_{i}', 0, 100)
    root.add_categorical_param(
        'categorical', [str(k) for k in range(categorical_size)]
    )

    converter = pttac.PaddedTrialToArrayConverter.from_study_config(
        pyvizier.ProblemStatement(
            search_space=space,
            metric_information=[
                pyvizier.MetricInformation(
                    'x1', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE
                )
            ],
        ),
        padding_schedule=padding_schedule,
    )
    trials = []
    for i in range(num_trials):
      parameters = {}
      for j in range(num_continuous):
        parameters[f'double_{j}'] = 1.0 * j

      for j in range(num_integer):
        parameters[f'integer_{j}'] = j

      parameters['categorical'] = str(i % categorical_size)
      final_measurement = pyvizier.Measurement(steps=1, metrics={'y': 3 * i})
      trials.append(
          pyvizier.Trial(
              parameters=parameters, final_measurement=final_measurement
          )
      )

    features = converter.to_features(trials).padded_array

    self.assertSequenceEqual(
        features.shape, [expected_num_trials, expected_dimension]
    )

    labels = converter.to_labels(trials).padded_array
    self.assertSequenceEqual(labels.shape, [expected_num_trials, 1])

    features, labels = converter.to_xy(trials)
    features = features.padded_array
    labels = labels.padded_array
    self.assertSequenceEqual(
        features.shape, [expected_num_trials, expected_dimension]
    )
    self.assertSequenceEqual(labels.shape, [expected_num_trials, 1])

    recovered_parameters = converter.to_parameters(features)
    self.assertSequenceEqual(
        recovered_parameters[:num_trials], [t.parameters for t in trials]
    )


if __name__ == '__main__':
  absltest.main()
